/*****************************************************************
**
** Tcl Mail Extension for Windows
** RCS Version $Revision: 1.28 $
** RCS Last Change Date: $Date: 2002/03/27 18:51:59 $
** Version 1.0.1.5
** Original Author: Michael I. Schwartz, mschwart@nyx.net
** MSC merged code contribution: Chris Nelson, nelson@pinebush.com
** Read, Delete, Logon, Logoff code contribution: Mark Austin, maustin@anadigics.com
** 
** {LICENSE}
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR WITH PERMISSION TO USE, COPY, MODIFY, AND
** DISTRIBUTE IT FOR ANY PURPOSE WITH THE FOLLOWING CONDITIONS:
** 1) IN SOURCE FORM, THIS HEADER MUST BE PRESERVED AND THESE CONDITIONS PROPOGATED
** 2) IN BINARY FORM, THE ORIGINAL AUTHOR MUST BE ACKNOWLEDGED IN DOCUMENTATION AND
**    CREDITS SCREEN (IF ANY)
** 3) FOR ANY COMMERCIAL SALE, THE AUTHOR MUST BE NOTIFIED OF THE USE OF THIS CODE
**    IN A PARTICULAR PRODUCT; THE ORIGINAL AUTHOR MAY CITE 
**    THE COMMERCIAL SELLER AS A BENEFICIARY OF THE CODE.
** 
** IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT,
** INDIRECT, SPECIAL, INCIDENTAL,  OR CONSEQUENTIAL DAMAGES ARISING OUT OF
** THE USE OF THIS SOFTWARE, ITS DOCUMENTATION,  OR ANY DERIVATIVES
** THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
** 
** THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
** INCLUDING, BUT NOT LIMITED TO,  THE IMPLIED WARRANTIES OF
** MERCHANTABILITY,FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT. 
** THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND
** DISTRIBUTORS HAVE NO OBLIGATION  TO PROVIDE MAINTENANCE, SUPPORT,
** UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
** 
** {SYNOPSIS}
** 
** This file contains commands to extend TK for Windows 3.11, Windows 95,
** and Windows NT 4.0 features
** 
** The commands are:
** 
** mapimail
** 
** {MAIL}
** 
** mapimail send [-to tolist] [-subject subject] \
**           [-body text] [-attachments attachmentlist]
**  DESCRIPTION:
**    Uses cmc_send_documents() or MAPISendMail() to send mail.
**  OPTIONS:
**    -to
**      tolist is a tcl list consisting of mail addresses
**      Each address must be acceptable to your mailer; aliases
**      are supported.
**      If the name starts with cc:, the remainder is placed as
**      a carbon-copy recipient.
**      If the name starts with bcc:, the remainder is placed as
**      a blind carbon-copy recipient (assuming bcc is supported
**      by your mailer).
**      If no -to argument is given, the mail program's dialog is invoked.
**    -subject
**      Goes into the subject of the message
**    If no -subject is provided, the subject is "No Subject"
**    -body
**      The text supplied here goes into the body of the message
**      If no -body is provided, there is no message body
**    -attachments
**      attachmentlist is a tcl list of filenames to be attached to the message.
**      If no -attachments is provided, there are no attachments.
**      tolist and attachmentlist are tcl lists.
**    -debug
**      This is a TEMPORARY option. Using this, the mail program will
**      ALWAYS bring up its dialog.
**  LIMITATIONS:
**   Under Win32s (Win3.11) MSMail does not appear to support MAPI32.
**   Eudora does, but the mail program must be loaded or a MAPI32.DLL
**   NOT FOUND dialog will be displayed.
**   If you run multiple mailers, your installation options determine
**   which mailer is invoked by this mail program.
**
** mapimail address [-clear]
**  DESCRIPTION:
**    Uses MAPIAddress or cmc_address_list to get a list of recipients
**    formatted for use in the mail send command.
**  OPTIONS:
**    -clear  clears the current recipient list before proceeding
**  LIMITATIONS:
**    Works only under Win95 and WinNT
**
** mapimail enabled [-detail]
**  DESCRIPTION:
**   returns 1 if MAPI is enabled in the Windows registry, 0 otherwise
**  OPTIONS:
**   -detail  Provides the version info in MAPI32.DLL if found
**
** mapimail logon [-user username] [-password passwd] [-ui true|false]
**  DESCRIPTION:
**   Logs onto a (new) mail session
**   If no options are provided, tries to logon with the "null" profile.
**  OPTIONS:
**   -user
**      Specifies the username or profile name for logon
**   -password
**      If specifying a user, the password for the user is provided.
**      With a legal user and password, the UI will not need to come up.
**   -ui
**      If true (default), the UI will come up if needed.
**  LIMITATIONS:
**
** mapimail logoff
**  DESCRIPTION:
**   Logs out of any mail session started with mail logon.
**  LIMITATIONS:
**
** mapimail option [-mailer cmc|mapi] [-verbose n]
**  DESCRIPTION:
**   Sets any available extension-level options.
**  OPTIONS:
**    -mailer
**      Determines whether CMC or MAPI mail protocols are used.
**      CMC is simpler, but more MS mailers support MAPI.
**      For instance, earlier versions of Eudora support MAPI only.
**      The default, therefore, is mapi.
**    -verbose
**      Specifies the level of messages to return from the mail extension.
**      The default, 0, has the minimum messages.
**  LIMITATIONS:
**    Only delete currently reports extra information if verbose.
**
** mapimail read [-digest] [-peek] [-attachments true|false]
**           [-bodyfile true|false] [-reset] -new|-all
**  DESCRIPTION:
**   Gets the contents of the "next" message in the mailbox.
**  OPTIONS:
**   -new (not default)
**     If specified, only looks at new messages
**   -all (not default)
**    If specified, looks at all messages, beginning with the first message
**   -reset (not default)
**    If specified, restarts message counter.
**   -digest (not default)
**    Gets the "envelope" information without getting the body of the message
**   -peek (not default)
**    Gets the next message without marking it as read
**   -attachments (true)
**    Specifies whether to retrieve the attachments or not.
**   -bodyfile (false)
**    Specifies whether the textual body of the message should be treated as a
**    file (or as a string)
**  LIMITATIONS:
**   Supported interfaces don't match between CMC and MAPI, so some differences in the
**   return values exist. For instance MAPI returns the time received, while CMC
**   returns the time sent.
**   Only supports "inbox" without using extensions.
**
** mapimail delete messageid ...
**   Removes the named messages from the mailbox.
**  OPTIONS:
**  LIMITATIONS:
**
** mapimail version
**  DESCRIPTION:
**    Returns the version of the mail command
**  LIMITATIONS:
**    None.
**
****************************************************************==*/
/************************REVISION HISTORY***************************
** Please see MAPIMail.log for further entries
**
** Revision 1.16  1998/10/20  04:30:01  Michael_Schwartz
** Recheck in due to RCS error
**
** Revision 1.16  1998/10/20  04:12:37  Michael_Schwartz
** Broke out some of the larger functions in the read routine
**
** Revision 1.15  1998/10/20  02:11:34  Michael_Schwartz
** Added all the error checking conditions for tkmsmail read commands.
** Added a -id to the read subcommand so that there is a way to follow
** up the read -digest subcommand.
**
** Revision 1.14  1998/10/08  06:31:36  Michael_Schwartz
** Cleaned up comments to ready for release
**
** Revision 1.13  1998/10/08  06:30:21  Michael_Schwartz
** Moved debug code to bottom and protected with #ifdef DEBUG.
**
** Revision 1.12  1998/10/08  06:21:24  Michael_Schwartz
** Updated the CMC portions of the code to give it a reasonable chance of working
**
** Revision 1.11  1998/09/29  03:25:02  Michael_Schwartz
** Updated after sending code out for review
**
** Revision 1.10  1998/09/28  02:17:50  Michael_Schwartz
** Fixed an indirection problem in the CMC code.
** Made logon interoperate with send and read properly for MAPI.
** Supported Originator field correctly.
** Removed all compilation warnings except those related to Sun-recommended
** macros.
**
** Revision 1.9  1998/09/27  09:00:45  Michael_Schwartz
** Updated to add Mark Austin's code for logon, logoff, read, and delete.
** Also modified to add the "mapimail option" command and to deprecate the
** -mailer arguments for all other commands.
** The problem of understanding and managing attachment files when mapi copies
** them is NOT yet managed, nor are all the additional details of the recipient
** list returned in the digest (i.e., it is not quite in a form usable by
** send, and does not distinguish between To, CC, and BCC recipients).
** Added CMC equivalents, but not yet tested.
**
** Revision 1.8  1998/09/23  03:38:13  Michael_Schwartz
** Updated for TclBlast effort.
** Support for MSVC compiler
**
** Revision 1.7  1998/04/23  03:53:03  Michael_Schwartz
** Added common code provided by Chris Nelson to finish up common baseline
** for MSC and Borland
**
** Revision 1.6  1997/12/17  06:35:35  Michael_Schwartz
** Update for version 0.6.0.2, includes 'mapimail address'
**
** Revision 1.5  1997/11/19  05:52:11  Michael_Schwartz
** *** empty log message ***
**
** Revision 1.4  1997/05/30  06:04:30  Michael_Schwartz
** Updated comments and usage message for source release version 0.5.0.2
**
** Revision 1.3  97/01/13  23:58:54  Michael_I_Schwartz
** version 0.5.0.1 binary release
** 
*****************************************************************/

#include <windows.h>
#include <io.h> /* For access() */
#include <direct.h> /* for getcwd() */
#include <tcl.h>
#include <xcmc.h>    /* This is the only file I could find cmc prototypes in... */
#include <xcmcext.h> /* In case we use any cmc extensions.... */
#include <xcmcmsxt.h>
#include <mapi.h>    /* This is for the mapi interface.. .*/
#include <stdlib.h>

#if defined(__WIN32__) || defined (__WIN32S__) || defined (WIN32S)
  #   ifdef EXPORT
  #       undef EXPORT
  #   endif
  #   if defined(_MSC_VER)
  #     include "smapi.h"       /* specs for initialization routine used by MSC */
  #     if TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION < 3 )
  #         define EXPORT(a,b) __declspec(dllexport) a b
  #         ifdef EXTERN
  #             undef EXTERN
  #             define EXTERN
  #         endif
  #     else
  #         define EXPORT(a,b) a b
  #     endif
  #     define DllEntryPoint DllMain
  #     define strncmpi strnicmp
  #   else
  #     if defined(__BORLANDC__)
  #         define EXPORT(a,b) a _export b
  #         ifdef EXTERN
  #             undef EXTERN
  #             define EXTERN
  #         endif
  #     else
  #         define EXPORT(a,b) a b
  #     endif
  #   endif
#else
  #   error "This extension is only for Windows. Sorry!"
#endif

/* New macros for tcl8.0.3 and later */
#if defined(TCL_STORAGE_CLASS)
#  undef TCL_STORAGE_CLASS
#endif
#define TCL_STORAGE_CLASS DLLEXPORT
#if ! defined(EXTERN)
#  define EXTERN
#endif

/****************************************************************
** External function prototypes
****************************************************************/
EXTERN EXPORT(int,Mapimail_Init) (Tcl_Interp *inter);

/****************************************************************
** Internal function prototypes
****************************************************************/

static int SendDocument ( Tcl_Interp *interp, int argc, char **argv);
static int Addresses    ( Tcl_Interp *interp, int argc, char **argv);
static int Version      ( Tcl_Interp *interp, int argc, char **argv);
static int Logon        ( Tcl_Interp *interp, int argc, char **argv);
static int Logoff       ( Tcl_Interp *interp, int argc, char **argv);
static int Read         ( Tcl_Interp *interp, int argc, char **argv);
static int Delete       ( Tcl_Interp *interp, int argc, char **argv);
static int Option       ( Tcl_Interp *interp, int argc, char **argv);
static int Enabled      ( Tcl_Interp *interp, int argc, char **argv);
static void FormatCmcError (CMC_return_code code, char *buffer, int maxsize);
static void FormatMapiError (long code, char *buffer, int maxsize);

/* Routines used to reduce complexity of especially long commands */
static long DefaultMapiLogon (void);
static long NextMapiMessage( FLAGS search_flags );
static void FormatMapiResult ( Tcl_Interp *interp, MapiMessage *m,
                               Tcl_DString *poutput, int responsibility );
static void FormatCmcDigestResult  ( Tcl_Interp *interp, CMC_message_summary FAR *pmsgref,
                                     Tcl_DString *poutput );
static void FormatCmcCompleteResult ( Tcl_Interp *interp, CMC_message_summary FAR *pmsgref,
                                      CMC_message FAR *msg, Tcl_DString *poutput);

/* Structure used in a utility routine */
struct msgref
{
  CMC_uint32 length;
  char       string[256+1];
};

/* Structure used to support mail functionality profile */
static struct capability
{
  const char *name;
  int         present;
} Mail_Capabilities [] =
{
  { "cmc support",             0 },
  { "MAPI support",            0 },
  { "send support",            0 },
  { "read support",            0 },
  { "delete support",          0 },

  { "MAPILogon",               0 },
  { "MAPILogoff",              0 },
  { "MAPISendMail",            0 },
  { "MAPISendDocuments",       0 },
  { "MAPIFindNext",            0 },
  { "MAPIReadMail",            0 },
  { "MAPISaveMail",            0 },
  { "MAPIDeleteMail",          0 },
  { "MAPIFreeBuffer",          0 },
  { "cmc_free",                0 },
  { "cmc_send_documents",      0 },
  { "cmc_look_up",             0 },
  { "cmc_logon",               0 },
  { "cmc_query_configuration", 0 },
  { "cmc_logoff",              0 },
  { "cmc_read",                0 },
  { "cmc_list",                0 },
  { "cmc_act_on",              0 },
};

/* Utility Routines */
static char * CmcMsgRefToString(char *cp, int *pmaxlen);
static int CmcStringToMsgRef(const char *string, struct msgref *pmsgref);
static HWND GetTopLevelHWND(void);
static void ProfileMailFunctions(void);
static int  get_mail_function(const char *string);
static long WinVersion(void);

/* Debug routines */
static void IAmHere(Tcl_Interp *interp, char *message);
static void DumpMsgRef(struct msgref *pmsgref);

/* Constants */
static char version_string[] = "1.0.2.1";
static char initial_directory[1024];

  /* Create a global structure to hold all data that must be manipulated when
  ** the program exits.
  */
  static struct globalstructs
  {
    CMC_recipient FAR *oldRecips;
    lpMapiRecipDesc    oldRecipients;
  } mail_data;

/****************
** Set defined constants for the mailer
** Use MAPI as the default.
****************/
#define USE_CMC 0
#define USE_MAPI 1
#define USE_DEFAULT USE_MAPI

#define USE_DEFAULT_STRING "MAPI"

/****************
** Global structures for options
** Perhaps all global structs should migrate to be indexed by
** the interp, so that this extension can become multithreaded.
** But is simple mapi multithreaded?
****************/
static struct optionstruct
{
  char default_mailer[8];
  int debug_level;
} optionstruct =
{ 
  USE_DEFAULT_STRING,
  0
};

/****************
** Global structure for stored info
** (see above note on implementation)
****************/
static struct sessionstruct
{
  CMC_session_id cmc_session;
  struct msgref  cmc_msgref;
  LHANDLE        mapi_session;
  char           msgId[512];
} sessionstruct =
{
  0,
  { 0, "" },
  0,
  ""
};

#if TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION <= 5
  /* In this case, must replace Tcl_Alloc() and Tcl_Free()
  ** with things based in Tcl_Ckalloc() and Tcl_Ckfree()
  */
#define Tcl_Alloc(x)  Tcl_Ckalloc(x)
#define Tcl_Free(x)   Tcl_Ckfree(x)

#endif

/* Exported symbols */
BOOL APIENTRY DllEntryPoint (HINSTANCE hInstance, DWORD reason, LPVOID lpCmdLine)
{
  switch (reason)
  {
    case DLL_PROCESS_ATTACH:
#if defined(_MSC_VER)
        /* Bind all the procs to a global address */
	if (!InitSimpleMAPI()) {
            return FALSE; /* FAIL */
	}
#endif
      /* Remember the filename for version info information */
      GetModuleFileName(hInstance, initial_directory, sizeof initial_directory );
      /* set the rightmost \\ to a null to remove the filename component of the directory */
      {
          char *cp;
          if ( (cp = strrchr(initial_directory, '\\')) != 0 )
              *cp = '\0';
      }
      
      /* Check out which mail functions are available so we can behave properly */
      ProfileMailFunctions();
      break;
    case DLL_THREAD_ATTACH:
      break;
    case DLL_PROCESS_DETACH:
      /* Release all the memory allocated by the recipient lists */
      if ( WinVersion() != VER_PLATFORM_WIN32s )
      {
        if ( mail_data.oldRecips )
        {
          cmc_free(mail_data.oldRecips);
          mail_data.oldRecips = 0;
        }
        if ( mail_data.oldRecipients )
        {
          if ( MAPIFreeBuffer )
            MAPIFreeBuffer(mail_data.oldRecipients);
          mail_data.oldRecipients = 0;
        }
      }
      break;
    case DLL_THREAD_DETACH:
      break;
  }
  /* Don't do anything, so just return true */
  return TRUE;
}

/****************************************************************
** Main mail command.
** Each subcommand strips itself off before calling its processor
****************************************************************/
EXPORT(int,Mail) (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  static char *usgmsg = "Usage:\n  "
                        "mapimail send "
                                  "[-to tolist] "
                                  "[-subject subject] [-body text] "
		                  "[-attachments attachmentlist] "
			          "[-debug]\n  "
			"mapimail address [-clear] \n  "
			"mapimail logon   [-user username] "
			             "[-password passwd] [-ui true|false]\n  "
			"mapimail logoff\n  "
			"mapimail option [-mailer cmc|mapi] [-verbose n]\n  "
                        "mapimail read [-digest] [-peek] "
                                  "[-attachments true|false] [-bodyfile true|false] "
				  "[-reset] -new|-all \n  "
			"mapimail delete messageid\n  "
                        "mapimail enabled [-detail]\n  "
			"mapimail version";
  argc--;
  argv++;
  /* Now the "mail" part is stripped off */
  
  if ( argc == 0 )
  {
    Tcl_SetResult(interp, usgmsg, TCL_STATIC);
    return TCL_ERROR;
  }

  if ( strcmp(argv[0], "send") == 0 )
  {
    argc--;
    argv++;
    return SendDocument(interp, argc, argv);
  }
  else if ( strcmp(argv[0], "address") == 0 )
  {
    argc--;
    argv++;
    return Addresses(interp, argc, argv);
  }
  else if ( strcmp(argv[0], "version") == 0 )
  {
    argc--;
    argv++;
    return Version(interp, argc, argv);
  }
  else if ( strcmp(argv[0], "logon") == 0 )
  {
    argc--;
    argv++;
    return Logon(interp, argc, argv);
  }
  else if ( strcmp(argv[0], "logoff") == 0 )
  {
    argc--;
    argv++;
    return Logoff(interp, argc, argv);
  }
  else if ( strcmp(argv[0], "read") == 0 )
  {
    argc--;
    argv++;
    return Read(interp, argc, argv);
  }
  else if ( strcmp(argv[0], "delete") == 0 )
  {
    argc--;
    argv++;
    return Delete(interp, argc, argv);
  }
  else if ( strcmp(argv[0], "option") == 0 )
  {
    argc--;
    argv++;
    return Option(interp, argc, argv);
  }
  else if ( strcmp(argv[0], "enabled") == 0 )
  {
    argc--;
    argv++;
    return Enabled(interp, argc, argv);
  }
  /* Unhandled option */
  Tcl_SetResult(interp, usgmsg, TCL_STATIC);
  return TCL_ERROR;  
}

/****************************************************************
** To support arbitrary mailers, this function should always
** return a "reasonable" value. Thus, leaving off the argument
** or providing a non-working value under, say, Unix, will still
** lead to usually-desirable results.
** Setting the default for the program is left to the higher-level
** application for now--though perhaps it ought to go in the
** registry.
****************************************************************/
static int StringToMailer (const char *string)
{
  int retval = USE_DEFAULT; /* Whichever that is.... */
  if (string == 0)
    return retval;
    
  switch (string[0])
  {
    case 'c': case 'C':
      retval = USE_CMC;
      break;
    case 'm': case 'M':
      retval = USE_MAPI;
      break;
  }
  return retval;
}

/****************************************************************
** Even though cmc is supposed to be more portable,
** it seems that Eudora, among others, supports only Simple MAPI.
****************************************************************/
static int SendDocument ( Tcl_Interp *interp, int argc, char **argv)
{
  int i;
  char *illegal = 0;
  char *tolist = 0;
  char *subject = "No subject";
  char *body = "Empty body";
  char *attachments = 0;
  char error_buf[200+1];

  /* Default flags */
  CMC_flags cmcflags = CMC_LOGON_UI_ALLOWED;
  long MAPI_flags = MAPI_LOGON_UI;
  
  static char cmcdelim[] = "^";
  static char usage[]    =  "Usage: mapimail send "
                            "[-to tolist] [-subject subject] "
                            "[-body text] [-attachments attachmentlist]\n"
		            "where tolist and attachmentlist are Tcl lists";

  
  int debug = 0;
  char *mailer = optionstruct.default_mailer;
  int mailer_type;  

#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION >= 1 )
   char *subjs, *bodys;
   Tcl_DString tds;
   Tcl_DString tdb;
   Tcl_DString tdt;
   Tcl_DString tda;
#endif

  /* To get results portably in and out of the interp's result, need a Tcl_DString 
  ** For Tcl 8.0 and later, direct use of interp's result is deprecated
  */
  Tcl_DString tmpstr;

  Tcl_DStringInit(&tmpstr);
  
  for (i=0; i<argc; i++)
  {
    if (strcmp(argv[i], "-to") == 0)
      tolist = argv[++i];     /* tolist is a Tcl list */
    else if ( strcmp(argv[i], "-subject") == 0)
      subject = argv[++i];    /* subject is a string */
    else if ( strcmp(argv[i], "-body") == 0)
      body = argv[++i];       /* body is a string */
    else if ( strcmp(argv[i], "-attachments") == 0)
      attachments = argv[++i]; /* attachments is a Tcl list */
    else if ( strcmp(argv[i], "-mailer") == 0 )
      mailer = argv[++i];
    else if ( strcmp(argv[i], "-debug") == 0 )
    {
      /*
      ** If debug is invoked, we both print verbosely and add confirmation
      ** dialogs. In the Unix version, both these things happen as well.
      */
      debug++;
      cmcflags |= CMC_ERROR_UI_ALLOWED | CMC_SEND_UI_REQUESTED;
      MAPI_flags |= MAPI_DIALOG;
    }
    else
      illegal = argv[i];
  }

  if (illegal || argc == 0)
  {
    Tcl_SetResult(interp, usage, TCL_STATIC);
    return TCL_ERROR;
  }

  /* If there are no addressees, we must use the mail application's dialogs
  ** to provide them. In the Unix version, a dialog pops up to request
  ** the missing portion.
  */
  if (tolist == 0)
  {
    cmcflags |= CMC_SEND_UI_REQUESTED;
    MAPI_flags |= MAPI_DIALOG;
  }
  
  mailer_type = StringToMailer(mailer);
  
#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION >= 1 )
  Tcl_DStringInit(&tds);
  if (subject)
  {
    Tcl_UtfToExternalDString(NULL, subject, -1, &tds);
    subjs = Tcl_DStringValue(&tds);
  }
  else
    subjs = subject;

  Tcl_DStringInit(&tdb);
  if (body)
  {
    Tcl_UtfToExternalDString(NULL, body, -1, &tdb);
    bodys = Tcl_DStringValue(&tdb);
  }
  else
    bodys = body;

  Tcl_DStringInit(&tdt);
  Tcl_DStringInit(&tda);
#endif

  if ( mailer_type == USE_CMC )
  {
    char *recipients = 0; /* tolist after replacement of cmcdelim */
    char *filepaths  = 0; /* attachments after replacement of cmcdelim */
  
    CMC_return_code retcode;

    /* CMC, fortunately, is a lot like Tcl--using string lists with a known
    ** separator. For now, use a fixed separator; in the future, it would be
    ** nice to have a set of handy separators, and use one of the "unused" ones
    ** to be the "real" separator
    */

    if ( get_mail_function("cmc support") == 0 || get_mail_function("send support") == 0 )
    {
      Tcl_SetResult(interp, "Can't support this operation in CMC mailer", TCL_STATIC);
      return TCL_ERROR;
    }

    if (tolist == 0)
      cmcflags |= CMC_SEND_UI_REQUESTED;
    else if ( Tcl_VarEval(interp, "join {", tolist, "} ", cmcdelim, NULL) == TCL_OK)
    {
#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION >= 1 )
      Tcl_DStringGetResult(interp, &tdt);
      Tcl_UtfToExternalDString(NULL, Tcl_DStringValue(&tdt), Tcl_DStringLength(&tdt), &tmpstr);
      Tcl_DStringFree(&tdt);
#else
      Tcl_DStringGetResult(interp, &tmpstr);
#endif
      if ( (recipients = Tcl_Alloc(Tcl_DStringLength(&tmpstr) + 1)) != 0 )
        strcpy(recipients, Tcl_DStringValue(&tmpstr));
      Tcl_DStringFree(&tmpstr);
    }
  
    if ( attachments )
    {
      if ( Tcl_VarEval(interp, "join {", attachments, "} ", cmcdelim, NULL) == TCL_OK)
      {
#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION >= 1 )
        Tcl_DStringGetResult(interp, &tda);
        Tcl_UtfToExternalDString(NULL, Tcl_DStringValue(&tda), Tcl_DStringLength(&tda), &tmpstr);
        Tcl_DStringFree(&tda);
#else
        Tcl_DStringGetResult(interp, &tmpstr);
#endif
        if ( (filepaths = Tcl_Alloc(Tcl_DStringLength(&tmpstr) + 1)) != 0)
          strcpy(filepaths, Tcl_DStringValue(&tmpstr));
        Tcl_DStringFree(&tmpstr);
      }
    }
  
    if (debug)
    {
      char *r = recipients?recipients:"NONE";
      char *s = subject?subject:"NONE";
      char *b = body?body:"NONE";
      char *f = filepaths?filepaths:"NONE";
      
      Tcl_VarEval (interp, "puts stdout { ",
                   "Sending CMC mail with:",
  		   "\nRecipients: ", r,
                   "\nSubject: ", s,
                   "\nBody: ", b,
                   "\nAttachments: ", f,
                   "\nDelimiter: ", cmcdelim,
                   " }", NULL);
    }

    /* Almost too easy, n'est pas? */
#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION >= 1 )
    retcode = cmc_send_documents(recipients, subjs, bodys,
                                 cmcflags, 
			         filepaths, 
			         filepaths, 
			         cmcdelim, 
			         0);
    Tcl_DStringFree(&tds);
    Tcl_DStringFree(&tdb);
#else
    retcode = cmc_send_documents(recipients, subject, body,
                                 cmcflags, 
			         filepaths, 
			         filepaths, 
			         cmcdelim, 
			         0);
#endif

    /* The cmc_send_documents mechanism starts a session (preferably without a UI)
    ** sends the documents and closes the session.
    */
    
    if (recipients)
      Tcl_Free(recipients);
    if (filepaths)
      Tcl_Free(filepaths);

    if (retcode != 0)
    {
      FormatCmcError(retcode, error_buf, 200);
      Tcl_SetResult(interp, error_buf, TCL_VOLATILE);
      return TCL_ERROR;
    }
  }
  else
  {
    /* For Simple MAPI, must put everything into appropriate structures */
    /* MapiSendDocuments is easier to use, but unfortunately, does not allow
    ** use of subject, body, and address lists.
    */
    /* To do this, we must parse each list. Sigh. */
    MapiMessage m;
    long retcode;
    
    char *recipfree = 0;
    char *attachfree = 0;
    char *scp;
    
    Tcl_DString *tos = 0;
    Tcl_DString *tat = 0;

    if ( get_mail_function("MAPI support") == 0 || get_mail_function("send support") == 0 )
    {
      Tcl_SetResult(interp, "Can't support this operation in MAPI mailer", TCL_STATIC);
      return TCL_ERROR;
    }

    memset(&m, 0, sizeof(MapiMessage));

#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION >= 1 )
    m.lpszSubject = subjs;
    m.lpszNoteText = bodys;
#else
    m.lpszSubject = subject;
    m.lpszNoteText = body;
#endif

    /* Ensure that no newlines remain in the subject.
    ** MAPI barfs if there are.
    */
    if (m.lpszSubject)
      while ( (scp = strchr(m.lpszSubject, '\n')) != 0 )
        *scp = ' ';
      
    /* Work through the recipients, if any */
    if (tolist)
    {
      int argc;
      char **argv;
      
      /* Split the list */
      if ( Tcl_SplitList(interp, tolist, &argc, &argv) != TCL_OK )
        return TCL_ERROR;
      
      if ( (m.lpRecips = (MapiRecipDesc *)Tcl_Alloc(argc * sizeof (MapiRecipDesc))) != 0 )
      {
        int i;
#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION >= 1)
        tos = (Tcl_DString *)Tcl_Alloc(argc *sizeof(Tcl_DString));

        if (tos)
        {
          int j;
          for (j=0;j<argc;j++)
            Tcl_DStringInit(&tos[j]);
        }
#endif

        m.nRecipCount = argc;
        memset (m.lpRecips, 0, argc * sizeof(MapiRecipDesc));
        /* Don't know whether lpszName or lpszAddress is proper here (or both!) */
        for (i=0; i<argc; i++)
        {
          if ( strncmpi(argv[i], "to:", 3) == 0)
          {
            if (tos)
            {
#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION >= 1)
              Tcl_UtfToExternalDString(NULL, argv[i] + 3, -1, &tos[i]);
#endif
              m.lpRecips[i].lpszName = m.lpRecips[i].lpszAddress = Tcl_DStringValue(&tos[i]);
            }
            else
              m.lpRecips[i].lpszName = m.lpRecips[i].lpszAddress = argv[i]+3;
            m.lpRecips[i].ulRecipClass = MAPI_TO;
          }
          else if ( strncmpi(argv[i], "cc:", 3) == 0 )
          {
            if (tos)
            {
#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION >= 1)
              Tcl_UtfToExternalDString(NULL, argv[i] + 3, -1, &tos[i]);
#endif
              m.lpRecips[i].lpszName = m.lpRecips[i].lpszAddress = Tcl_DStringValue(&tos[i]);
            }
            else
              m.lpRecips[i].lpszName = m.lpRecips[i].lpszAddress = argv[i]+3;
            m.lpRecips[i].ulRecipClass = MAPI_CC;            
          }
          else if ( strncmpi(argv[i], "bcc:", 4) == 0)
          {
            if (tos)
            {
#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION >= 1)
              Tcl_UtfToExternalDString(NULL, argv[i] + 4, -1, &tos[i]);
#endif
              m.lpRecips[i].lpszName = m.lpRecips[i].lpszAddress = Tcl_DStringValue(&tos[i]);
            }
            else
              m.lpRecips[i].lpszName = m.lpRecips[i].lpszAddress = argv[i]+4;
            m.lpRecips[i].ulRecipClass = MAPI_BCC;
          }
          else
          {
            if (tos)
            {
#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION >= 1)
              Tcl_UtfToExternalDString(NULL, argv[i], -1, &tos[i]);
#endif
              m.lpRecips[i].lpszName = m.lpRecips[i].lpszAddress = Tcl_DStringValue(&tos[i]);
            }
            else
              m.lpRecips[i].lpszName = m.lpRecips[i].lpszAddress = argv[i];
            m.lpRecips[i].ulRecipClass = MAPI_TO;
          }
        }
      }

      recipfree = (char *) argv;
    }
    
    if ( attachments )
    {
      int argc;
      char **argv;
      
      /* Split the list */
      if ( Tcl_SplitList(interp, attachments, &argc, &argv) != TCL_OK )
      {
        if (recipfree)
          Tcl_Free(recipfree);
        return TCL_ERROR;
      }

      attachfree = (char *)argv;
      
      if ( (m.lpFiles = (MapiFileDesc *)Tcl_Alloc(argc * sizeof(MapiFileDesc))) != 0 )
      {
        int i;
#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION >= 1)
        tat = (Tcl_DString *)Tcl_Alloc(argc * sizeof(Tcl_DString));

        if (tat)
        {
          int j;
          for (j=0; j<argc; j++)
            Tcl_DStringInit(&tat[j]);
        }
#endif
        memset(m.lpFiles, 0, argc*sizeof(MapiFileDesc));
        m.nFileCount = argc;

        for (i=0; i<argc; i++)
        {
          m.lpFiles[i].nPosition = -1;
          if ( tat )
          {
#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION >= 1)
            Tcl_UtfToExternalDString(NULL, argv[i], -1, &tat[i]);
#endif
            m.lpFiles[i].lpszPathName = m.lpFiles[i].lpszFileName = Tcl_DStringValue(&tat[i]);
          }
          else
            m.lpFiles[i].lpszPathName = m.lpFiles[i].lpszFileName = argv[i];
        }
      }
    }

    Tcl_ResetResult(interp); /* Lose the return values of the other Tcl calls. */
    if (debug)
    {
      char *t = tolist?tolist:"No tolist";
      char *s = subject?subject:"No subject";
      char *b = body?body:"No body";
      char *a = attachments?attachments:"No attachments";
      unsigned i;

      /* The puts is for immediate feedback; the result provides additional detail */
      Tcl_VarEval (interp, "puts stdout { ",
                            "Sending MAPI mail with:",
			    "\nRecipients: ", t,
			    "\nSubject: ", s,
			    "\nBody: ", b,
			    "\nAttachments: ", a,
			    " }", NULL);

      Tcl_ResetResult(interp);
      Tcl_AppendElement(interp, "Addressees:");
      for (i=0; i< m.nRecipCount; i++)
        Tcl_AppendElement(interp, m.lpRecips[i].lpszAddress);
      Tcl_AppendElement(interp, "Attachments:");
      for (i=0; i< m.nFileCount; i++)
        Tcl_AppendElement(interp, m.lpFiles[i].lpszFileName);
    }

    /* Find a valid MAPI session or fail, requiring logon */
    if ( sessionstruct.mapi_session == 0 )
    {
      /* Establish the default session if possible */
      if ( (retcode = DefaultMapiLogon() ) != SUCCESS_SUCCESS )
      {
        Tcl_SetResult(interp, 
	              "Cannot establish default logon to mail.\n"
	              "Please use mapimail logon before mapimail send", TCL_STATIC);
        return TCL_ERROR;
      }
      else if ( optionstruct.debug_level > 0 )
        Tcl_AppendResult(interp, "Logged into MAPI with Default profile\n", NULL);
    }
    
    /* Now call the function. This is what is in the .h; 
    ** not, unfortunately, what is in the  MAPI documentation
    */
    retcode = MAPISendMail (
                  sessionstruct.mapi_session,  /* session */
		  (ULONG)GetTopLevelHWND(),    /* UI param */
		  &m,                          /* Message */
		  MAPI_flags,                  /* flags */
		  0                            /* reserved */
                );

    /* Free up MAPI-specific memory */
    if (recipfree)
      Tcl_Free (recipfree);
    if (attachfree)
      Tcl_Free (attachfree);
#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION >= 1 )
    Tcl_DStringFree(&tds);
    Tcl_DStringFree(&tdb);
    if (tos)
    {
      int j;
      for (j=0;(unsigned int)j<m.nRecipCount;j++)
        Tcl_DStringFree(&tos[j]);
      Tcl_Free((char *)tos);
    }
    if (tat)
    {
      int j;
      for (j=0;(unsigned int)j<m.nFileCount;j++)
        Tcl_DStringFree(&tat[j]);
      Tcl_Free((char *)tat);
    }
#endif

    if (retcode != SUCCESS_SUCCESS)
    {
      FormatMapiError(retcode, error_buf, 200);
      Tcl_SetResult(interp, error_buf, TCL_VOLATILE);
      return TCL_ERROR;
    }
  }
  
  return TCL_OK;
}

/****************************************************************
** Addresses
** Return list of addresses formatted for use with SendMail
****************************************************************/
static int Addresses    ( Tcl_Interp *interp, int argc, char **argv)
{
  /* Default flags */
  CMC_flags cmcflags = CMC_LOOKUP_ADDRESSING_UI | CMC_LOGON_UI_ALLOWED | CMC_ERROR_UI_ALLOWED | CMC_LOOKUP_RESOLVE_PREFIX_SEARCH;
  long MAPI_flags    = MAPI_LOGON_UI;
  
  ULONG nEditFields = 4;
  char *mailer = optionstruct.default_mailer;
  int mailer_type;
  int i;
  char buffer[200+1]; /* Long enough for any "reasonable" recipient (?) */
  int do_nameonly = 0;
  int do_clear    = 0;

  /* Exit early if this is Windows 3.1 */
  if ( WinVersion() == VER_PLATFORM_WIN32s )
  {
    Tcl_SetResult(interp, "mapimail address: Not implemented in Win32s", TCL_STATIC);
    return TCL_ERROR;
  }

  /* Handle Args */
  for (i=0; i<argc; i++)
  {
    if ( strcmp(argv[i], "-mailer") == 0 )
      mailer = argv[++i];
    else if ( strcmp(argv[i], "-clear" ) == 0 ) /* Clear the current list */
    {
      do_clear = 1;
    }
    else if ( strcmp(argv[i], "-name") == 0 ) /* Use only the name for MAPI */
      do_nameonly = 1;
  }
  
  /* Get to work! */
  mailer_type = StringToMailer(mailer);
  
  if ( mailer_type == USE_CMC )
  {
    CMC_return_code status;
    CMC_uint32      count = 0;
    CMC_uint32      k;
    CMC_recipient FAR *newRecips;

    if ( do_clear && mail_data.oldRecips )
    {
        cmc_free(mail_data.oldRecips);
        mail_data.oldRecips = 0;
    }

    if ( (unsigned long)cmc_look_up == 0 )
    {
      Tcl_SetResult(interp, "Can't support this operation in CMC mailer", TCL_STATIC);
      return TCL_ERROR;
    }

    status = cmc_look_up (sessionstruct.cmc_session,    /* session */
                          mail_data.oldRecips,          /* Pre-populated recipients */
			  cmcflags,                     /* Default flags */
			  (CMC_ui_id)GetTopLevelHWND(), /* UI Id */
			  &count,                       /* receive count from output */
			  &newRecips,                   /* Resulting recipients */
			  0);                           /* CMC Extensions */
    if (status != 0)
    {
      FormatCmcError(status, buffer, 200);
      Tcl_SetResult(interp, buffer, TCL_VOLATILE);
      return TCL_ERROR;
    }

    if (mail_data.oldRecips)
      cmc_free(mail_data.oldRecips);

    mail_data.oldRecips = newRecips;
    newRecips = 0;

    for (k=0; k< count; k++ )
    {
      char *cl = "to";
      switch (mail_data.oldRecips[k].role)
      {
        case CMC_ROLE_TO:     cl = "to";    break;
	case CMC_ROLE_CC:     cl = "cc";    break;
	case CMC_ROLE_BCC:    cl = "bcc";   break;
      }
      sprintf(buffer, "%s:%s", cl, mail_data.oldRecips[k].name);
      Tcl_AppendElement(interp, buffer);
    }
  }
  else /* MAPI */
  {
    ULONG status;
    static ULONG nold, nnew;
    static lpMapiRecipDesc newRecips;
    ULONG k;

    if ( do_clear && mail_data.oldRecipients )
    {
      MAPIFreeBuffer(mail_data.oldRecipients);
      mail_data.oldRecipients = 0;
      nold = 0;
    }

    if ( (unsigned long)MAPIAddress == 0 )
    {
      Tcl_SetResult(interp, "Can't use addressing functions in MAPI with this mailer",
                    TCL_STATIC);
      return TCL_ERROR;
    }
    
    status = MAPIAddress (0,                        /* session */
                          (ULONG)GetTopLevelHWND(), /* UI id */
			  "Addresses",              /* Caption */
			  nEditFields,              /* Number of recipient classes permitted */
			  0,                        /* Labels for edit controls (NULL requested) */
			  nold,                     /* Number of oldRecips */
			  mail_data.oldRecipients,  /* Pre-populated recipients */
			  MAPI_flags,               /* Default flags */
			  0,                        /* Reserved */
			  &nnew,                    /* Receive number of new recipients */
			  &newRecips);              /* pointer to array of new recipients */
    if (status != SUCCESS_SUCCESS)
    {
      FormatMapiError(status, buffer, 200);
      Tcl_SetResult(interp, buffer, TCL_VOLATILE);
      return TCL_ERROR;
    }

    if (mail_data.oldRecipients)
      MAPIFreeBuffer(mail_data.oldRecipients);

    nold = nnew;
    mail_data.oldRecipients = newRecips;
    nnew = 0;
    newRecips = 0;

    {
      Tcl_DString rslt;
      Tcl_DStringInit(&rslt);
      
      for (k=0; k<nold; k++)
      {
        char *cl="to";
        switch (mail_data.oldRecipients[k].ulRecipClass)
        {
          case 1: cl = "to";      break;
          case 2: cl = "cc";      break;
          case 3: cl = "bcc";     break;
        }
        if ( do_nameonly == 0 &&
	     mail_data.oldRecipients[k].lpszAddress && 
             mail_data.oldRecipients[k].lpszAddress[0])
        {
          /* Handle Eudora MAPI quoting bug, reported 12/2/1999
          ** Note: This fix is not perfectly general (it doesn't fix
          **       the recipient list itself), but handles
          **       the most common situation (a single pass at the address list).
          **       Hopefully, Eudora will fix the
          **       bug before a more general solution becomes necessary.
          */
          if ( mail_data.oldRecipients[k].lpszAddress[0] == '"' &&
               strrchr(mail_data.oldRecipients[k].lpszAddress, '"') ==
                 mail_data.oldRecipients[k].lpszAddress         &&
               k+1 < nold )
            sprintf(buffer, "%s:%s, %s", cl, 
	                    mail_data.oldRecipients[k].lpszAddress,
                            mail_data.oldRecipients[1+k++].lpszAddress);
          else if ( mail_data.oldRecipients[k].lpszAddress[0] == '\\' &&
               k+1 < nold )
            sprintf(buffer, "%s:%s, %s", cl, 
	                    mail_data.oldRecipients[k].lpszAddress+1,
                            mail_data.oldRecipients[1+k++].lpszAddress);
          else
            sprintf(buffer, "%s:%s", cl, mail_data.oldRecipients[k].lpszAddress);
        }
        else
          sprintf(buffer, "%s:%s", cl, mail_data.oldRecipients[k].lpszName);
        Tcl_DStringAppendElement(&rslt, buffer);
      }
      Tcl_DStringResult(interp, &rslt);
    }
  }

  return TCL_OK;
}

/****************************************************************
** Logon
** Create Session Using Specified Profile With Mailer
** Needs option to allow login with or without UI. If UI is
** not requested, should it be provided anyway if the login fails?
** UI should not be provided iff profile and password available?
** Needs option to use existing shared session(?)
** Must support -mailer flag to support either CMC or MAPI.
****************************************************************/
static int Logon      ( Tcl_Interp *interp, int argc, char **argv)
{
  /* Local Variables */
  char msg[64];
  long retcode;
  char *mailer = optionstruct.default_mailer;
  int  mailer_type;
  static char usage[] = "mapimail logon [-user username] [-password passwd] "
                                    "[-ui true|false]";
  char *username = 0;
  char *password = 0;
  int  use_ui = 0;
  int  i;
  FLAGS mapi_flags = MAPI_NEW_SESSION;
  CMC_flags cmc_flags = CMC_ERROR_UI_ALLOWED;
  char error_buf[200+1];

  HWND parent = 0;
  
  /* If mapi_session is non-zero this has already been done */
  if ( sessionstruct.mapi_session != 0 )
  {
    Tcl_SetResult (interp, "Warning: Logon after login is completed", TCL_STATIC);
    return TCL_OK;
  }

  /* Handle the arguments */
  for (i=0; i<argc; i++)
  {
    if ( strcmp(argv[i], "-user") == 0 || strcmp(argv[i], "-profile") == 0 )
      username = argv[++i];
    else if ( strcmp(argv[i], "-password") == 0 )
      password = argv[++i];
    else if ( strcmp(argv[i], "-mailer") == 0 )
      mailer = argv[++i];
    else if ( strcmp(argv[i], "-ui") == 0 )
    {
      if ( argv[++i] == 0 )
          use_ui = 1;
      else	
        switch (argv[i][0])
	{
          case '0': case 'f': case 'F': /* False or 0 (99+% of the time) */
            use_ui = 0;
            break;
          default:  /* True or 1 */
            use_ui = 1;
            break;
        }
    }
    else
    {
      Tcl_SetResult(interp, usage, TCL_STATIC);
      return TCL_ERROR;
    }
  }

  /* Check if there's enough information to do without a UI.
  ** If not, just allow one silently.
  */
  if ( username == 0 || password == 0 || use_ui != 0 )
  {
     cmc_flags  |= CMC_LOGON_UI_ALLOWED;
     mapi_flags |= MAPI_LOGON_UI;
     parent      = GetTopLevelHWND();
  }
  
  mailer_type = StringToMailer(mailer);
  
  /* Initialize Message Id For Reading */
  sessionstruct.msgId[0] ='\0';

  if ( mailer_type == USE_MAPI )
  {
    if ( get_mail_function("MAPI support") == 0 )
    {
      Tcl_SetResult(interp, "No MAPI support in this mailer", TCL_STATIC);
      return TCL_ERROR;
    }
    
    /* Create New Session To Mailer Using Profile Name */
    retcode = MAPILogon((ULONG)parent,                    /* Parent window (modal) or 0 */
                        username,                         /* Profile name */
		        password,                         /* Password */
		        mapi_flags,                       /* Flags */
		        0,                                /* reserved--must be 0 */
		        &sessionstruct.mapi_session);     /* Mail session handle */

    if ( retcode != SUCCESS_SUCCESS )
    {
      FormatMapiError(retcode, error_buf, 200);
      Tcl_SetResult(interp, error_buf, TCL_VOLATILE);
      return TCL_ERROR;
    }
  }
  else /* CMC */
  {
    CMC_return_code retcode;
    CMC_object_identifier charset;  /* Use cmc_query_configuration() to get */
    CMC_uint16 cmc_version;
    CMC_boolean ui_avail;
    CMC_X_COM_support sup[2];
    CMC_extension ext;

    if ( get_mail_function("cmc support") == 0 )
    {
      Tcl_SetResult(interp, "CMC unsupported in this mailer", TCL_STATIC);
      return TCL_ERROR;
    }
    
    retcode = cmc_query_configuration((CMC_session_id)0,      /* Session (none yet) */
                                      CMC_CONFIG_UI_AVAIL,    /* Item */
                                      (CMC_buffer) &ui_avail, /* Result */
                                      NULL);                  /* CMC Extensions */
    /* If no UI is available, the implementation is somewhere between
       lame and nonexistent. Let the user know
    */
    if ( retcode != 0 )
    {
      FormatCmcError(retcode, error_buf, 200);
      Tcl_AppendResult (interp, "CMC appears unsupported: ", error_buf, "\nTry MAPI", NULL);
      return TCL_ERROR;
    }
    
    retcode = cmc_query_configuration(0,          /* Session (none yet) */
                                      CMC_CONFIG_CHARACTER_SET,   /* Item */
				      (CMC_buffer) &charset,      /* Result buffer */
				      NULL);      /* CMC Extensions */

    if ( retcode != 0 )
    {
      /* Can't get character set--leave it a 0 (default)
      ** Emit warning if verbosity is set high enough
      */
    }
    
    retcode = cmc_query_configuration(0,          /* Session (none yet) */
                                      CMC_CONFIG_VER_SPEC,        /* Item */
                                      (CMC_buffer) &cmc_version,  /* Result buffer */
                                      NULL);      /* CMC Extensions */

    if ( retcode != 0 )
    {
      /* Can't get the cmc version--leave it at 0 (default)
      ** Emit warning if verbosity is set high enough
      */
    }
    
    /* Now set up the CMC extensions */
    sup[0].item_code = CMC_XS_COM;
    sup[0].flags     = 0;
    sup[1].item_code = CMC_XS_MS;
    sup[1].flags     = 0;
    ext.item_code = CMC_X_COM_SUPPORT_EXT;
    ext.item_data = 2;
    ext.item_reference = sup;
    ext.extension_flags = CMC_EXT_LAST_ELEMENT;
    
    retcode = cmc_logon (NULL,                    /* User interface */
                         username,                /* Username/profile */
			 password,                /* Password */
			 (CMC_enum)0,             /* Character set  (charset) */
			 (CMC_ui_id)parent,       /* Parent UI id */
			 cmc_version,             /* Client CMC version * 100 */
			 cmc_flags,               /* Flags */
			 &sessionstruct.cmc_session,  /* CMC Session */
			 &ext );                  /* CMC extensions (e.g., MAPI) */

    if ( retcode != 0 )
    {
      FormatCmcError(retcode, error_buf, 200);
      Tcl_AppendResult(interp, "CMC logon failed: ", error_buf, "\nTry MAPI", NULL);
      return TCL_ERROR;
    }
  }
  
  /* Form Return Message */
  if ( username != 0 )
    sprintf(msg, "Logged On Using %s Profile", username);
  else if ( use_ui != 0 )
    sprintf(msg, "Logged On Using UI-selected Profile");
  else
    strcpy(msg, "Logged On Using Default Profile");
  Tcl_AppendElement(interp, msg);
  
  /* OK Return */
  return TCL_OK;
}

/****************************************************************
** LogOff
** Mailer System
** Here, we won't ask for any parameters(?)
** Just log off of any open session.
****************************************************************/
static int Logoff      ( Tcl_Interp *interp, int argc, char **argv)
{
  /* Local Variables */
  long retcode;
  char error_buf[200+1];

  if ( sessionstruct.mapi_session != 0 )
  {
    /* Logoff and Delete Current Session  */
    retcode = MAPILogoff(sessionstruct.mapi_session,0,0,0);

    if ( retcode != SUCCESS_SUCCESS )
    {
      FormatMapiError(retcode, error_buf, 200);
      Tcl_AppendResult(interp, error_buf, "\n", 0);
    }
    else
      Tcl_AppendResult(interp, "Logged off MAPI session\n", 0);

    /* Initialize Mail Session To Zero Again, So It will Ask If Needed */
    sessionstruct.mapi_session = 0;
    sessionstruct.msgId[0] = '\0';
  }

  if ( sessionstruct.cmc_session != 0 )
  {
    CMC_return_code retcode;

    retcode = cmc_logoff(sessionstruct.cmc_session, 
                         0,
 			 CMC_LOGOFF_UI_ALLOWED | CMC_ERROR_UI_ALLOWED , 
			 0);

    sessionstruct.cmc_session = 0;
    sessionstruct.cmc_msgref.length = 0;
    sessionstruct.cmc_msgref.string[0] = '\0';
    
    if ( retcode != 0 )
    {
      FormatCmcError(retcode, error_buf, 200);
      Tcl_AppendResult(interp, error_buf, "\n", 0);
    }
    else
    {
      Tcl_AppendResult(interp, "Logged off CMC session\n", 0);
    }
  }
  
  Tcl_AppendResult(interp, "Logged Off Mail System", 0);
  
  /* OK Return */
  return TCL_OK;
}

/****************************************************************
** Read
** Read Next Message From Mail System
** Should it allow a range of messages to be read? All new messages?
****************************************************************/
static int Read ( Tcl_Interp *interp, int argc, char **argv)
{
  /* Local Variables */
  char           error_buf[200];
  MapiMessage    *m;
  FLAGS          mapi_flags = 0;
  FLAGS          mapi_srch_flags = 0;
  CMC_flags      cmc_flags  = 0;
  static char usage[] = "mapimail read [-digest] [-peek] "
                                  "[-attachments true|false] [-bodyfile true|false] "
				  "[-new|-all] [-reset|-id messageid]";

  char *mailer = optionstruct.default_mailer;
  int  mailer_type;
  int  do_content = 1;
  int  do_peek    = 0;
  int  ignore_attachments = 1;
  int  body_as_attachment = 0;
  int  do_new = 0;
  int  do_all = 0;
  int  do_reset = 0;
  int  do_msgid = 0;
  const char *msgid = 0;
  
  int i;

  Tcl_DString output;

  Tcl_DStringInit(&output);
  

  if ( get_mail_function("read support") == 0 )
  {
    Tcl_SetResult(interp, "No read support in this mailer", TCL_STATIC);
    return TCL_ERROR;
  }
  
  /* Process the argument list */
  for (i=0; i<argc; i++)
  {
    if ( strcmp(argv[i], "-mailer") == 0 )
      mailer = argv[++i];
    else if ( strcmp(argv[i], "-digest") == 0 )
      do_content = 0;
    else if ( strcmp(argv[i], "-peek") == 0 )
      do_peek = 1;
    else if ( strcmp(argv[i], "-attachments") == 0 )
    {
      if ( argv[++i] == 0 )
      {
        Tcl_SetResult(interp, usage, TCL_STATIC);
        return TCL_ERROR;
      }
      switch ( argv[i][0] )
      {
        case '0': case 'f': case 'F': /* False */
          ignore_attachments = 1 ;
          break;
        default:
          ignore_attachments = 0;
          break;
      }
    }
    else if ( strcmp(argv[i], "-bodyfile") == 0 )
    {
      if ( argv[++i] == 0 )
      {
        Tcl_SetResult(interp, usage, TCL_STATIC);
        return TCL_ERROR;
      }
      switch ( argv[i][0] )
      {
        case '0': case 'f': case 'F': /* False */
          body_as_attachment = 0 ;
          break;
        default:
          body_as_attachment = 1;
          break;
      }
    }
    else if ( strcmp(argv[i], "-new") == 0 )
      do_new = 1;
    else if ( strcmp(argv[i], "-all") == 0 )
      do_all = 1;
    else if ( strcmp(argv[i], "-reset") == 0 )
      do_reset = 1;
    else if ( strcmp(argv[i], "-id") == 0 )
    {
      do_msgid = 1;
      msgid = argv[++i];
    }
    else
    {
      Tcl_SetResult(interp, usage, TCL_STATIC);
      return TCL_ERROR;
    }
  }

  /* Check if we have a valid session */
  mailer_type = StringToMailer(mailer);
  if ( mailer_type == USE_CMC  && sessionstruct.cmc_session  == 0 )
  {
    Tcl_SetResult(interp, "mapimail logon must precede mapimail read", TCL_STATIC);
    return TCL_ERROR;
  }
  else if  (mailer_type == USE_MAPI && sessionstruct.mapi_session == 0 )
  {
    if ( DefaultMapiLogon() != SUCCESS_SUCCESS )
    {
      Tcl_SetResult(interp, 
                    "Cannot establish default logon to mail.\n"
	            "Use mapimail logon before mapimail read", TCL_STATIC);
        return TCL_ERROR;
    }
    else if ( optionstruct.debug_level > 0 )
    {
      Tcl_DStringStartSublist(&output);
      Tcl_DStringAppendElement(&output, "Diagnostics");
      Tcl_DStringAppendElement(&output, "Logged into MAPI with Default profile");
      Tcl_DStringEndSublist(&output);
      /* Set the result just in case there is an early return */
      Tcl_SetResult(interp, "Logged into MAPI with Default profile", TCL_STATIC);
    }
  }

  /* Check Got Correct Number Of Arguments */
  if ( do_new + do_all + do_msgid != 1 )
  {
    Tcl_AppendResult(interp, "Error: Must specify one of -new or -all or -id\n", usage, NULL);
    return TCL_ERROR;
  }
  /* Check whether our options are sensible */
  if ( do_msgid && do_reset )
  {
    Tcl_AppendResult(interp, "Error: Can't have both -reset and -id\n", usage, 0);
    return TCL_ERROR;
  }

  if ( mailer_type == USE_MAPI )
  {	
    long retcode;

    if ( get_mail_function("MAPI support") == 0 )
    {
      Tcl_AppendResult(interp, "Error: No MAPI support in this mailer\n", 0 );
      return TCL_ERROR;
    }
    
    /* Set up the mapi flags right */
    if ( do_msgid )
    {
      if ( msgid && msgid[0] )
        strcpy(sessionstruct.msgId, msgid);
      else /* Illegal id--reset and report error if verbosity on */
      {
        sessionstruct.msgId[0] = '\0';
        if ( optionstruct.debug_level > 0 )
        {
          Tcl_DStringStartSublist(&output);
          Tcl_DStringAppendElement(&output, "Error");
          Tcl_DStringAppendElement(&output, "Illegal message id--resetting mail");
          Tcl_DStringEndSublist(&output);
        }
      }
    }
    else if ( do_reset )
      sessionstruct.msgId[0] = '\0';
    if ( ignore_attachments != 0 )
      mapi_flags |= MAPI_SUPPRESS_ATTACH; /* If OFF, creates a temp file! */
    if ( do_content == 0 )
      mapi_flags |= MAPI_ENVELOPE_ONLY;   /* If ON,  creates a temp file! */
    if ( do_peek )
      mapi_flags |= MAPI_PEEK;
    if ( body_as_attachment != 0 )
      mapi_flags |= MAPI_BODY_AS_FILE;    /* If ON,  creates a temp file! */

    if ( do_new )
      mapi_srch_flags |= MAPI_UNREAD_ONLY;

    if ( do_msgid == 0 )
    {
      /* Check If All Messages or New Or Reset Position */
      retcode = NextMapiMessage(mapi_srch_flags);

      /* Check For No Mail */
      if ( retcode == MAPI_E_NO_MESSAGES ) 
      {
        Tcl_AppendResult(interp,"No Messages",NULL);
        return TCL_OK;
      }

      /* Check For Success */
      if ( retcode != SUCCESS_SUCCESS ) 
      {
        FormatMapiError(retcode, error_buf, 200);
        Tcl_AppendResult(interp, error_buf, TCL_VOLATILE);
        return TCL_ERROR;
      }
    }
    
    /* Read The Mail Message */
    retcode = MAPIReadMail(sessionstruct.mapi_session,  /* Session */
                           (ULONG)GetTopLevelHWND(),    /* UI parameter */
			   sessionstruct.msgId,         /* message id */
			   mapi_flags,                  /* Flags */
			   0,                           /* Reserved - must be 0 */
			   &m);                         /* MapiMessage */

    if ( retcode != SUCCESS_SUCCESS )
    {
      /* Is this always serious? */
      FormatMapiError(retcode, error_buf, 200);
      switch (retcode)
      {
        case MAPI_E_ATTACHMENT_WRITE_FAILURE: /* Not too bad--just no attachment */
        case MAPI_E_DISK_FULL:                /* Not too bad--just no attachments */
        case MAPI_E_TOO_MANY_FILES:           /* Not too bad--limited attachments */
        case MAPI_E_TOO_MANY_RECIPIENTS:      /* Not too bad--just limited recipients */
          if (optionstruct.debug_level > 0 )
          {
            Tcl_DStringStartSublist(&output);
            Tcl_DStringAppendElement(&output, "Error");
            Tcl_DStringAppendElement(&output, error_buf);
            Tcl_DStringEndSublist(&output);
          }
          break;
        case MAPI_E_FAILURE: case MAPI_E_INSUFFICIENT_MEMORY:
        case MAPI_E_INVALID_MESSAGE: case MAPI_E_INVALID_SESSION:
        default: /* Anything unexpected */
          /* These are all bad enough to quit about */
          Tcl_AppendResult(interp, error_buf, TCL_VOLATILE);
          if (m)
            MAPIFreeBuffer(m);
          return TCL_ERROR;
      }
    }
    FormatMapiResult (interp, m, &output, (mapi_flags & MAPI_ENVELOPE_ONLY) ||
                                          (mapi_flags & MAPI_SUPPRESS_ATTACH ) );

    /* Put Into TCL Return Structure */
    Tcl_DStringResult(interp, &output);

    Tcl_DStringFree(&output);

    /* Free Message Stucture Up */
    if (m)
    {
      MAPIFreeBuffer(m);
      m = 0;
    }
  }
  else /* CMC */
  {
    CMC_return_code retcode;
    CMC_message_summary  FAR *pmsgref = NULL;
    CMC_message_reference FAR *pseed   = (CMC_message_reference FAR *)&sessionstruct.cmc_msgref;
    CMC_message_reference FAR *pread;
    CMC_message FAR *msg = 0;
    CMC_uint32 msgrefcnt = 1; /* We want at most 1 message */
    struct msgref tmpmsgref;

    if ( get_mail_function("cmc support") == 0 )
    {
      Tcl_AppendResult(interp, "Error: ", "cmc not supported by this mailer\n", 0 );
      return TCL_ERROR;
    }
    
    /* Set up the flags right */
    cmc_flags |= CMC_ERROR_UI_ALLOWED ;

    if ( do_reset || sessionstruct.cmc_msgref.length == 0)
      pseed = 0;
    else if ( do_msgid )
    {
      if ( CmcStringToMsgRef(msgid, &tmpmsgref) != 0 )
      {
        Tcl_AppendResult(interp, "Error: ", "Invalid format for CMC msg id\n", 0);
        return TCL_ERROR;
      }
      else
      {
        pseed = (CMC_message_reference FAR *)&tmpmsgref;
        pread = (CMC_message_reference FAR *)&tmpmsgref;
      }
    }

    if ( do_new )
      cmc_flags |= CMC_LIST_UNREAD_ONLY;

    if ( do_msgid == 0 )
    {
      /* In order to do what we want in CMC, we must first cmc_list to get the
      ** reference we want, and then cmc_read that message reference.
      ** The previous reference can be used as a seed to start from.
      */
      retcode = cmc_list ( sessionstruct.cmc_session,   /* CMC session */
                           NULL,                        /* Message type */
                           cmc_flags,                   /* Flags */
                           pseed,                       /* Pointer to msg ref seed */
                           &msgrefcnt,                  /* Count of messages to do */
			   (CMC_ui_id)GetTopLevelHWND(), /* UI Id */
                           &pmsgref,                    /* Result array */
                           (CMC_extension FAR *) 0);    /* Extensions */

      if (retcode != 0 )
      {
        FormatCmcError(retcode, error_buf, 200);
        Tcl_AppendResult(interp, error_buf, NULL);
        return TCL_ERROR;
      }

      if ( msgrefcnt == 0 )   /* no more messages */
      {
        Tcl_AppendResult(interp, "No Messages", NULL);
        return TCL_OK;
      }

      /* Setup the reference. Handle references up to 256 bytes only */
      if ( pmsgref != 0 && pmsgref[0].message_reference != 0 )
      {
        if ( (sessionstruct.cmc_msgref.length = pmsgref[0].message_reference->length) > 256 )
          sessionstruct.cmc_msgref.length = 256;
        memcpy(sessionstruct.cmc_msgref.string, pmsgref[0].message_reference->string,
                 sessionstruct.cmc_msgref.length);
        sessionstruct.cmc_msgref.string[sessionstruct.cmc_msgref.length] = '\0';
      }
      pread = pmsgref->message_reference;
    }
    else
    {
      if ( CmcStringToMsgRef ( msgid, &sessionstruct.cmc_msgref) != 0 )
      {
        Tcl_SetResult(interp, "Error: Illegal CMC message reference structure", TCL_STATIC);
        return TCL_ERROR;
      }
    }
    
    /* If only a digest is needed, we already have everything necessary! */
    if ( do_content == 0 && do_msgid == 0 )
    {
      FormatCmcDigestResult(interp, pmsgref, &output);
      
      Tcl_DStringResult(interp, &output);
  
      Tcl_DStringFree(&output);
    }
    else /* CMC Mailer with -id or not with -digest */
    {
      char idstring[2*24+3];
      char *pidstring;
      int  size = sizeof(idstring);

#if DEBUG
      /* Debug */
      DumpMsgRef(&sessionstruct.cmc_msgref);
#endif
      
      if ( (pidstring = CmcMsgRefToString(idstring, &size)) == 0 )
        pidstring = "Unknown";
        
      cmc_flags = CMC_ERROR_UI_ALLOWED;
      if ( ignore_attachments != 0 )
        cmc_flags |= CMC_MSG_AND_ATT_HDRS_ONLY;   /* If OFF, creates a temp file! */
      if ( do_peek )
        cmc_flags |= CMC_DO_NOT_MARK_AS_READ;

      retcode = cmc_read ( sessionstruct.cmc_session,     /* CMC session */
                           pread,                         /* CMC_message_reference */
                           cmc_flags,                     /* Flags */
			   &msg,                          /* Message pointer address */
			   (CMC_ui_id)GetTopLevelHWND(),  /* UI Id */
			   (CMC_extension FAR *) 0);      /* extensions */
      if ( retcode != 0 )
      {
        FormatCmcError(retcode, error_buf, 200);
        Tcl_SetResult(interp, error_buf, TCL_VOLATILE);
        return TCL_ERROR;
      }

      /* OK. Now set up the output structure */
      /* Ensure the parts exist before copying them! */
      Tcl_DStringStartSublist(&output);
      Tcl_DStringAppendElement(&output, "id");
      Tcl_DStringAppendElement(&output, pidstring);
      Tcl_DStringEndSublist(&output);

      FormatCmcCompleteResult (interp, pmsgref, msg, &output);

      /* According to the MAPI pages, we must free the recipients now... 
      ** Oddly enough, no such statement is made for the attachments.
      */
      if ( msg->recipients )
        cmc_free(msg->recipients);

      /* Assume we need to free the attachments */
      if ( msg->attachments )
        cmc_free(msg->attachments);

      Tcl_DStringResult(interp, &output);
  
      Tcl_DStringFree(&output);

      /* We have to free any leftover stuff now */
      /* Certainly the msg structure itself... */
      if (msg)
        cmc_free(msg);
    }
    /* Free the "list" structure */
    if (pmsgref)
      cmc_free(pmsgref);
  }

  /* OK Return */
  return TCL_OK;
}

/****************************************************************
** Delete
** Delete Message(s) From Mail System
****************************************************************/
static int Delete ( Tcl_Interp *interp, int argc, char **argv)
{
  /* Local Variables */
  static char usage[] = "mapimail delete message_id ...";
  int i;
  char *mailer = optionstruct.default_mailer;
  int  mailer_type;
  char error_buf[200];

  if ( get_mail_function("delete support") == 0 )
  {
    Tcl_SetResult(interp, "Can't delete mail with this mail support DLL", TCL_STATIC);
    return TCL_ERROR;
  }
  
  /* Process arguments */
  for (i=0; i<argc && argv[i][0] == '-'; i++)
  {
    if ( strcmp(argv[i], "-mailer") == 0 )
      mailer = argv[++i];
    else
    {
      Tcl_SetResult(interp, usage, TCL_STATIC);
      return TCL_ERROR;
    }
  }
  
  /* If missing any ids, raise an error */
  if ( i >= argc ) 
  {
    Tcl_SetResult(interp, usage, TCL_STATIC);
    return TCL_ERROR;
  }

  mailer_type = StringToMailer(mailer);

  if ( mailer_type == USE_MAPI)
  {
    long  retcode;
    
    if ( (unsigned long)MAPIDeleteMail == 0 )
    {
      Tcl_SetResult(interp, "Can't delete mail using MAPI", TCL_STATIC);
      return TCL_ERROR;
    }
    
    for ( ; i < argc; i++ )
    {
      /* Delete The Mail Message */
      retcode = MAPIDeleteMail(sessionstruct.mapi_session,0,argv[i],0,0);
      if ( retcode != SUCCESS_SUCCESS )
      {
        FormatMapiError(retcode,error_buf, 200);
        Tcl_AppendResult(interp, "Error: ", argv[i], ": ", error_buf, "\n", 0);
      }
      else if ( optionstruct.debug_level > 0 )
      {
        Tcl_AppendResult(interp, "Deleted ", argv[i], "\n", 0);
      }
    }
  }
  else /* CMC */
  {
    CMC_return_code retcode;
    CMC_X_COM_support sup[2];
    CMC_extension ext;

    /* Now set up the CMC extensions */
    sup[0].item_code = CMC_XS_COM;
    sup[0].flags     = 0;
    sup[1].item_code = CMC_XS_MS;
    sup[1].flags     = 0;
    ext.item_code = CMC_X_COM_SUPPORT_EXT;
    ext.item_data = 2;
    ext.item_reference = sup;
    ext.extension_flags = CMC_EXT_LAST_ELEMENT;

    if ( (unsigned long)cmc_act_on == 0 )
    {
      Tcl_SetResult(interp, "Can't delete mail with this CMC mail support DLL", TCL_STATIC);
      return TCL_ERROR;
    }
    
    for ( ; i < argc; i++ )
    {
      struct msgref tmpmsgref;

      /* Delete The Mail Message */
      /* First must get a valid reference, and then act_on with CMC_ACT_ON_DELETE */
      /* The message reference is the same as the message id */
      if ( CmcStringToMsgRef(argv[i], &tmpmsgref) != 0 )
      {
        Tcl_AppendResult(interp, "Error: ", "Invalid format for msg id\n", 0);
        return TCL_ERROR;
      }

#if DEBUG
      DumpMsgRef(&tmpmsgref);
#endif
      retcode = cmc_act_on(sessionstruct.cmc_session,     /* Session */
                           (CMC_message_reference *)&tmpmsgref,  /* Message reference */
			   CMC_ACT_ON_DELETE,             /* Operation */
                           0,                             /* act_on_flags */
                           (CMC_ui_id)GetTopLevelHWND(),  /* UI ID */
			   0);                            /* Extensions */
      if ( retcode != 0 )
      {
        FormatCmcError(retcode, error_buf, 200);
        Tcl_AppendResult(interp, "Error: ", argv[i], ": ", error_buf, "\n", 0);
      }
      else if ( optionstruct.debug_level > 0 )
      {
        Tcl_AppendResult(interp, "Deleted ", argv[i], "\n", 0);
      }
    }
  }
  
  /* OK Return */
  return TCL_OK;
}

/****************************************************************
** Option
** Set available extension-wide options
****************************************************************/
static int Option       ( Tcl_Interp *interp, int argc, char **argv)
{
  static char usage[] = "mapimail option [-mailer cmc|mapi] [-verbose n]";
  int i;
  int dousage = 0;
  
  if ( argc == 0 )
  {
    char verbosity[12];

    sprintf(verbosity, "%d", optionstruct.debug_level);
    Tcl_AppendResult(interp,
                     "{mailer ", optionstruct.default_mailer, "} "
		     "{verbose ", verbosity, "}", 
		     0);
    return TCL_OK;
  }
  
  /* Now process the args--an attempt to SET the options */
  for (i=0; i<argc; i++)
  {
    if (strcmp(argv[i], "-mailer") == 0 )
    {
      if ( ++i < argc )
      {
        strncpy(optionstruct.default_mailer,argv[i], sizeof(optionstruct.default_mailer));
        optionstruct.default_mailer[sizeof(optionstruct.default_mailer) - 1] = '\0';
      }
      else
        dousage = 1;
    }
    else if ( strcmp(argv[i], "-verbose") == 0 )
    {
      if ( ++i < argc )
      {
        char *endstr;
        int  val;

        val = strtol(argv[i], &endstr, 0);  /* Any valid base */
        if ( val >= 0 && endstr != argv[i] )
          optionstruct.debug_level = val;
        else
          dousage = 1;
      }
      else
        dousage = 1;
    }
    else
      break;
  }
  if ( i != argc || dousage )
  {
    Tcl_SetResult(interp, usage, TCL_STATIC);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/****************************************************************
 * For each directory, combine with the filename, and if it exists,
 * load up the inter result with the values in the version info
 ****************************************************************/
static int EnabledVersionFile(Tcl_Interp *interp, const char *filename, const char **dirs)
{
    /* 040904e4 is a language spec, a "magic number" from the VerQueryValueEx manual page */
    static char *info[] = {
        "\\StringFileInfo\\040904E4\\CompanyName", 
        "\\StringFileInfo\\040904E4\\FileDescription", 
        "\\StringFileInfo\\040904E4\\FileVersion", 
        "\\StringFileInfo\\040904E4\\InternalName",
        "\\StringFileInfo\\040904E4\\LegalCopyright", 
        "\\StringFileInfo\\040904E4\\OriginalFilename", 
        "\\StringFileInfo\\040904E4\\ProductName", 
        "\\StringFileInfo\\040904E4\\ProductVersion",
    };
    int infoskip = 25; /* sizeof(info[0]) - sizeof("CompanyName") */
    int i;
    int retval = TCL_OK;
    int found = 0;
    
    for (i=0; dirs[i] != 0 && found == 0; i++)
    {
        char *cp;
        if ( (cp = Tcl_Alloc(strlen(dirs[i])+strlen(filename)+1+1)) != 0 )
        {
            sprintf(cp,"%s/%s",dirs[i],filename);
            
            if ( access(cp,0) == 0 ) /* OK */
            {
                DWORD z;
                long siz;
                
                found = 1; /* This is the only one we'll look at */
                
                if ( (siz = GetFileVersionInfoSize(cp,&z)) != 0 )
                {
                    char *block;
                    
                    if ( (block = Tcl_Alloc(siz)) != 0 )
                    {
                        if ( GetFileVersionInfo(cp,0,siz,block) )
                        {
                            /* Now we can get to work reading the data! */
                            int j;
                            UINT len;
                            char *val;
                            Tcl_DString result;
                            
                            Tcl_DStringInit(&result);
                            
                            /* Add an element to represent what file is used */
                            Tcl_DStringStartSublist(&result);
                            Tcl_DStringAppendElement(&result, "FullPathName");
                            Tcl_DStringAppendElement(&result, cp);
                            Tcl_DStringEndSublist(&result);
                            
                            /* Get all the other info */
                            for (j=0; j<sizeof(info) / sizeof(char *); j++)
                            {
                                if ( VerQueryValue((LPVOID)block, info[j], (LPVOID *)&val, &len) )
                                {
                                    Tcl_DStringStartSublist(&result);
                                    Tcl_DStringAppendElement(&result, info[j]+infoskip);
                                    Tcl_DStringAppendElement(&result, val);
                                    Tcl_DStringEndSublist(&result);
                                }
                            }
                            Tcl_AppendResult(interp, Tcl_DStringValue(&result), 0);
                            Tcl_DStringFree(&result);
                        }
                        else /* Can't read the file version info data */
                        {
                        }
                        Tcl_Free(block);
                    }
                }
                else /* failed to get version info size */
                {
                }
            }
            else /* No access */
            {
            }
            Tcl_Free(cp);
        }
        else /* no memory */
        {
        }
    }
    return retval;
}

/****************************************************************
 * Check the registry for info about whether MAPI is loaded
 * HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Run\OptionalComponents\MAPI Installed
 * under windows will provide info on whether MAPI is installed
 * (What is the difference between this and 
 * "HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows Messaging Subsystem" MAPI 
 * ??)
 * Both keys return a null-terminated string, which should have one character: '0' or '1'
****************************************************************/
static int EnabledMAPIEnabled (Tcl_Interp *interp)
{
    /* Nothing in here requires the "fancy" versions of Open and Query */
    static const struct reg_qry {
        HKEY       hive;
        const char *key;
        const char *value;
    } key_info[] = {
        { HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run\\OptionalComponents\\MAPI", "Installed" },
        { HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\Windows Messaging Subsystem", "MAPI" },
    };
    HKEY key;
    int j;
    char buf[20];
    DWORD siz;
    DWORD pt;
    
    /* RegOpenKey */
    for (j = 0; j < sizeof (key_info ) / sizeof (struct reg_qry); j++ )
    {
        if ( RegOpenKeyEx(key_info[j].hive, key_info[j].key, 0, KEY_READ, &key) == ERROR_SUCCESS )
        {
            siz = sizeof buf;
            if ( RegQueryValueEx(key, key_info[j].value, NULL, &pt, buf, &siz) == ERROR_SUCCESS )
            {
                switch (buf[0])
                {
                case '0': /* Not enabled */
                    Tcl_AppendResult(interp, "0", 0);
                    break;
                case '1': /* Enabled */
                    Tcl_AppendResult(interp, "1", 0);
                    break;
                case 0:  /* A null character -- perhaps the result is in Unicode? */
                    Tcl_AppendResult(interp, buf+1, 0);
                    break;
                default: /* Unexpected */
                    Tcl_AppendResult(interp, buf, 0); /* Whatever it is, that's the result */ 
                    break;
                }
                break; /* Get out of the loop */
            }
        }
    }
    if ( j >= sizeof key_info / sizeof (struct reg_qry) )
    {
        Tcl_AppendResult(interp, "unknown", 0);
    }
    return TCL_OK;
}

/****************************************************************
 * Enabled
 * Return whether MAPI is enabled, and the details if available
 ****************************************************************/
static int Enabled      ( Tcl_Interp *interp, int argc, char **argv)
{
    int details = 0;
    char *dirs[4]; /* This directory and the windows directory, and end with a 0 */
    char curdir[1024];
    char sysdir[1024];
    char windir[1024];
    
    /* The order that (32-bit) DLLs are searched (per "LoadLibrary" man page)
     * a) Application location directory
     * b) Current directory
     * c) Windows system directory (i.e., SYSTEM32 subdirectory, GetSystemDirectory())
     * d) Windows directory (GetWindowsDirectory())
     * e) PATH
     * For the purposes of this application, (b), (c), and (d) are searched
     * If MAPI is properly installed, it is likely that (c) will always be the check that works
     */
    
    /* Trivial argument parsing */
    if ( argc > 0 )
    {
        if ( strncmp(argv[0], "-detail", 7) == 0 )
            details = 1;
    }
    
    EnabledMAPIEnabled(interp);
    
    if ( details )
    {
        int k = 0;
        
        Tcl_AppendResult(interp, "\n", 0);
        dirs[3] = 0;
        dirs[2] = 0;
        dirs[1] = 0;
        dirs[0] = 0;
        
        /* Don't you just love windows? Let's change the order of the arguments! */
        if ( GetCurrentDirectory(sizeof curdir, curdir) )
            dirs[k++] = curdir;
        
        if ( GetSystemDirectory(sysdir, sizeof sysdir) )
            dirs[k++] = sysdir;
        
        if ( GetWindowsDirectory(windir, sizeof windir) )
            dirs[k++] = windir;
        
        EnabledVersionFile(interp, "MAPI32.DLL", dirs);
    }
    return TCL_OK;
}

/****************************************************************
** Version
** Return version of the package
 * Looking at the version info in %WINDIR%\SYSTEM32\MAPI32.DLL will tell the exact data
 * (of course a MAPI32.DLL in the same directory as msmapi.dll will override the system verison...)
 * The query for the specific information uses the GetFileVersionInfo/VerQueryValue calls,
 * and these return unicode strings in all versions of windows.
****************************************************************/
static int Version      ( Tcl_Interp *interp, int argc, char **argv)
{
  Tcl_SetResult(interp, version_string, TCL_STATIC);
  return TCL_OK;
}

/****************************************************************
** GetTopLevelHWND
** This function gets a "reasonable" top level UI for attaching
** dialogs to. If it returns NULL, well, that's reasonable too!
****************************************************************/
static HWND GetTopLevelHWND(void)
{
  HWND toplevel;
  HWND tmp;
  if ( (toplevel = GetActiveWindow() ) != 0 )
    while ( (tmp = GetWindow(toplevel, GW_OWNER)) != 0 )
      toplevel = tmp;
  return toplevel; 
}

/*****************************************************************
** set_mail_function
****************************************************************/
static int set_mail_function( const char *name, unsigned long addr)
{
  int i;
  
  for (i = 0; i < sizeof (Mail_Capabilities) / sizeof (struct capability); i++)
  {
    if ( strcmp(Mail_Capabilities[i].name, name) == 0 )
    {
      if ( addr != 0L )
        Mail_Capabilities[i].present = 1;
      return 1;
    }
  }
  return 0;
}

/*****************************************************************
** get_mail_function
****************************************************************/
static int get_mail_function( const char *name)
{
  int i;
  
  for (i = 0; i < sizeof (Mail_Capabilities) / sizeof (struct capability); i++)
  {
    if ( strcmp(Mail_Capabilities[i].name, name) == 0 )
    {
      return  Mail_Capabilities[i].present;
    }
  }
  return 0;
}

/****************************************************************
** ProfileMailFunctions
**
** Possible high-level results are:
**   cmc support
**   MAPI support
**   send support
**   read support
**   insufficient support for mail
****************************************************************/
static void ProfileMailFunctions(void)
{
  set_mail_function ( "MAPILogon",               (unsigned long) MAPILogon);
  set_mail_function ( "MAPILogoff",              (unsigned long) MAPILogoff);
  set_mail_function ( "MAPISendMail",            (unsigned long) MAPISendMail);
  set_mail_function ( "MAPISendDocuments",       (unsigned long) MAPISendDocuments);
  set_mail_function ( "MAPIFindNext",            (unsigned long) MAPIFindNext);
  set_mail_function ( "MAPIReadMail",            (unsigned long) MAPIReadMail);
  set_mail_function ( "MAPISaveMail",            (unsigned long) MAPISaveMail);
  set_mail_function ( "MAPIDeleteMail",          (unsigned long) MAPIDeleteMail);
  set_mail_function ( "MAPIFreeBuffer",          (unsigned long) MAPIFreeBuffer);
  set_mail_function ( "cmc_free",                (unsigned long) cmc_free);
  set_mail_function ( "cmc_send_documents",      (unsigned long) cmc_send_documents);
  set_mail_function ( "cmc_look_up",             (unsigned long) cmc_look_up);
  set_mail_function ( "cmc_logon",               (unsigned long) cmc_logon);
  set_mail_function ( "cmc_query_configuration", (unsigned long) cmc_query_configuration);
  set_mail_function ( "cmc_logoff",              (unsigned long) cmc_logoff);
  set_mail_function ( "cmc_read",                (unsigned long) cmc_read);
  set_mail_function ( "cmc_list",                (unsigned long) cmc_list);
  set_mail_function ( "cmc_act_on",              (unsigned long) cmc_act_on);

  /* cmc support means that at least cmc_logon, cmc_logoff, cmc_query_configuration are present */
  if ( get_mail_function("cmc_logon") && 
       get_mail_function("cmc_logoff") &&
       get_mail_function("cmc_query_configuration") )
    set_mail_function("cmc support", 1 );
  /* MAPI support means that at least MAPILogon, MAPILogoff, and MAPISendDocuments are present */
  if ( get_mail_function("MAPILogon") && 
       get_mail_function("MAPILogoff") )
    set_mail_function("MAPI support", 1 );

  /* Send support means that either cmc_send_documents or MAPISendDocuments are supported */
  if ( get_mail_function("cmc_send_documents") ||
       get_mail_function("MAPISendDocuments") )
    set_mail_function("send support", 1);

  /* Read support means that either cmc_list, cmc_read, cmc_look_up, and cmc_act_on
   * are present or MAPIFindNext and MAPIReadmail are present
  */
  if ( ( get_mail_function("MAPIFindNext") &&
         get_mail_function("MAPIReadMail")
       )
       ||
       ( get_mail_function("cmc_list") &&
         get_mail_function("cmc_read") &&
         get_mail_function("cmc_look_up") &&
         get_mail_function("cmc_act_on")
       )
     )
    set_mail_function("read support", 1);

  /* Delete support means that either MAPIDeleteMail or cmc_act_on are present */
  if ( get_mail_function("MAPIDeleteMail") ||
       get_mail_function("cmc_act_on") )
    set_mail_function ("delete support", 1);   
}

/*****************************************************************
** FormatCmcError
*****************************************************************/
struct cmc_errmsg {
  CMC_return_code code;
  char *msg;
} cmc_error_msg[] =
{
  { 0, "Unexpected error code" },
  { CMC_E_ATTACHMENT_NOT_FOUND, "Attachment not found" },
  { CMC_E_ATTACHMENT_OPEN_FAILURE, "Can't open attachment" },
  { CMC_E_ATTACHMENT_READ_FAILURE, "Can't read attachment" },
  { CMC_E_ATTACHMENT_WRITE_FAILURE, "Can't write attachment" },
  { CMC_E_FAILURE, "Mail failure" },
  { CMC_E_INSUFFICIENT_MEMORY, "Mail failure due to lack of memory" },
  { CMC_E_INVALID_FLAG, "Mail failure due to bad flag" },
  { CMC_E_INVALID_PARAMETER, "Mail failure due to bad parameter"},
  { CMC_E_INVALID_MESSAGE_REFERENCE, "Invalid message reference" },
  { CMC_E_INVALID_UI_ID, "Mail failure due to invalid UI id" },
  { CMC_E_LOGON_FAILURE, "Unable to log into mail service" },
  { CMC_E_RECIPIENT_NOT_FOUND, "One or more recipients cannot be found " },
  { CMC_E_TEXT_TOO_LARGE, "Text size exceeded current implementation limit" },
  { CMC_E_TOO_MANY_FILES, "Number of attachments exceeded current implementation limit" },
  { CMC_E_TOO_MANY_RECIPIENTS, "Number of recipients exceeded current implementation limit" },
  { CMC_E_UNSUPPORTED_FLAG, "Mail failure due to unsupported flag" },
  { CMC_E_USER_CANCEL, "Mail delivery canceled" },
  { CMC_E_USER_NOT_LOGGED_ON, "User not logged into mail" },
  { CMC_E_AMBIGUOUS_RECIPIENT, "Ambiguous recipient--multiple matches found"},
  { CMC_E_INVALID_SESSION_ID, "Mail session id is invalid. Log into mail first."},
  { CMC_E_LOGON_FAILURE, "User name, password, or service invalid. Cannot log in."},
  { CMC_E_NOT_SUPPORTED, "This feature not supported in CMC." },
};

static void FormatCmcError (CMC_return_code code, char *buffer, int maxsize)
{
  int i;
  char numbuf[13]; /* long enough for the biggest long number */
  int len;
  sprintf(numbuf, "%ld ", code);
  len = strlen(numbuf);
  if (maxsize <= len)
  {
    return;
  }
  strcpy(buffer, numbuf);
  buffer += len;
  maxsize -= len;
  
  for (i=1; i< sizeof(cmc_error_msg) / sizeof(struct cmc_errmsg); i++)
  {
    if ( code == cmc_error_msg[i].code )
    {
      strncpy(buffer, cmc_error_msg[i].msg, maxsize);
      buffer[maxsize - 1] = '\0';
      return;
    }
  }
  strncpy (buffer, cmc_error_msg[0].msg, maxsize);
  buffer[maxsize - 1] = '\0';
}

/*****************************************************************
** FormatMapiError
*****************************************************************/
struct mapi_errmsg {
  long code;
  char *msg;
} mapi_error_msg[] =
{
  { 0                               , "Unexpected error code" },
  { MAPI_E_USER_ABORT               , "User aborted mail command" },
  { MAPI_E_FAILURE                  , "General mail failure" },
  { MAPI_E_LOGON_FAILURE            , "User mail login failure" },
  { MAPI_E_DISK_FULL                , "Mail failed due to disk full" },
  { MAPI_E_INSUFFICIENT_MEMORY      , "Mail failed due to insufficient memory" },
  { MAPI_E_ACCESS_DENIED            , "Mail failed due to permission problem" },
  { MAPI_E_TOO_MANY_SESSIONS        , "Too many mail sessions" },
  { MAPI_E_TOO_MANY_FILES           , "Number of attachment files exceeds current implementation limit" },
  { MAPI_E_TOO_MANY_RECIPIENTS      , "Number of recipients exceeds current implementation limit" },
  { MAPI_E_ATTACHMENT_NOT_FOUND     , "Attachment file not found" },
  { MAPI_E_ATTACHMENT_OPEN_FAILURE  , "Unable to open attachment file (share or permission problem)" },
  { MAPI_E_ATTACHMENT_WRITE_FAILURE , "Unable to write attachment in mail message" },
  { MAPI_E_UNKNOWN_RECIPIENT        , "Unknown recipient in mail message" },
  { MAPI_E_BAD_RECIPTYPE            , "Illegal recipient type in mail message" },
  { MAPI_E_INVALID_MESSAGE          , "Invalid mail message" },
  { MAPI_E_TEXT_TOO_LARGE           , "Text of mail message is too large" },
  { MAPI_E_INVALID_SESSION          , "Invalid mail session" },
  { MAPI_E_TYPE_NOT_SUPPORTED       , "Mail type not supported" },
  { MAPI_E_AMBIGUOUS_RECIPIENT      , "Ambiguous recipient in mail message" },
  { MAPI_E_INVALID_RECIPS           , "Invalid recipients in mail message" },
  { MAPI_E_NOT_SUPPORTED            , "MAPI is not supported" },
};

static void FormatMapiError (long code, char *buffer, int maxsize)
{
  int i;
  char numbuf[13]; /* long enough for the biggest long number */
  int len;
  sprintf(numbuf, "%ld ", code);
  len = strlen(numbuf);
  if (maxsize <= len)
  {
    return;
  }
  strcpy(buffer, numbuf);
  buffer += len;
  maxsize -= len;
  
  for (i=1; i< sizeof(mapi_error_msg) / sizeof(struct mapi_errmsg); i++)
  {
    if ( code == mapi_error_msg[i].code )
    {
      strncpy(buffer, mapi_error_msg[i].msg, maxsize);
      buffer[maxsize - 1] = '\0';
      return;
    }
  }
  strncpy (buffer, mapi_error_msg[0].msg, maxsize);
  buffer[maxsize - 1] = '\0';
}

/* Utility procedures */

/****************************************************************
** CmcMsgRefToString
** Converts a CMC message reference (ala microsoft, anyway)
** to a null-terminated string. MAPI does something like this
** for you, but CMC starts the references with some null bytes
**
** Input: pmaxlen is the maximum length of cp including the nul
** Output: returns 0 or cp; pmaxlen is set to the actual length copied
** Return:
**  0   Failure
**  cp  Success
****************************************************************/
static char * CmcMsgRefToString(char *cp, int *pmaxlen)
{
  char *s=cp;
  static char X[] = "0123456789ABCDEF";
  int maxlen = *pmaxlen;
  int len;
  
  if ( maxlen <= 0 )
  {
    *pmaxlen = 2 * sessionstruct.cmc_msgref.length + 3;
    return 0;
  }
  
  if (sessionstruct.cmc_msgref.length > 0 )
  {
    *cp++ = '0';
    *cp++ = 'x';
    for (len=0; (unsigned int)len < sessionstruct.cmc_msgref.length && 
                              len < maxlen - 1;
         len++)
    {
      *cp++ = X[(unsigned char)(sessionstruct.cmc_msgref.string[len]) >> 4];
      *cp++ = X[(unsigned char)(sessionstruct.cmc_msgref.string[len]) & 0xF];
    }
    *cp = '\0';
  }
  else
  {
    len = 0;
    *cp= '\0';
  }
  *pmaxlen = len;
  return s;
}

/****************************************************************
** CmcStringToMsgRef
** Undoes the effects of CmcMsgRefToString
** User supplies the msgref structure and string to convert.
**
** Returns
** 0  OK
** 1  string was NULL
** 2  pmsgref was NULL
** 3  An illegal byte (non 0-F) was found in string
**
****************************************************************/
/* Returns 0 on success, other on failure */
static int CmcStringToMsgRef(const char *string, struct msgref *pmsgref)
{
  const char *cp;
  int  i;
  int  len;
  char *s;

  const char X[] = "0123456789ABCDEF";
  
  /* If the string or msgref is null, return failure */
  if (string == 0 ) return 1;
  if (pmsgref == 0 ) return 2;

  len = strlen(string);
  s   = pmsgref->string;
  pmsgref->length = 0;
  cp  = string;
  
  /* If the string starts with 0x, skip over it. */
  if ( string[0] == '0' && (string[1] == 'X' || string[1] == 'x' ) )
  {
    cp  += 2;
    len -= 2;
  }

  /* Now, for each bit of string, add some characters to the reference */
  for ( i=0; i < len; i++)
  {
    int nibble1 = -1;
    int nibble2 = -1;
    int j;

    for (j=0; X[j]; j++)
      if (cp[i] == X[j])
      {
        nibble1 = j;
        break;
      }
    i++;
    for (j=0; X[j]; j++)
      if (cp[i] == X[j])
      {
        nibble2 = j;
        break;
      }

    if (nibble1 == -1 || nibble2 == -1 )
      return 3;

    *s++ = (nibble1 << 4) + nibble2;
    pmsgref->length++;
  }

  *s = '\0';
  return 0;
}

static long DefaultMapiLogon (void)
{
    return MAPILogon(0,                            /* Parent window (modal) or 0 */
                     0,                            /* Profile name */
                     0,                            /* Password */
	             0,                            /* Flags */
		     0,                            /* reserved--must be 0 */
		     &sessionstruct.mapi_session); /* Mail session handle */
}

static long NextMapiMessage( FLAGS search_flags )
{
  return  MAPIFindNext(sessionstruct.mapi_session,  /* Session */
                             (ULONG)GetTopLevelHWND(),    /* UI */
                             NULL,                        /* Message Type */
			     sessionstruct.msgId,         /* Message id seed */
			     search_flags,                /* Flags */
			     0,                           /* Reserved (0) */
			     sessionstruct.msgId);        /* Output message id */

}

static void FormatMapiResult ( Tcl_Interp *interp, MapiMessage *m, 
                               Tcl_DString *poutput, int responsibility )
{
  /* Put the result into a list of 2-element lists, key and value */
  /* Protect each section by ensuring the value is present! */
  if ( sessionstruct.msgId )
  {
    Tcl_DStringStartSublist(poutput);
    Tcl_DStringAppendElement(poutput, "id");
    Tcl_DStringAppendElement(poutput, sessionstruct.msgId);
    Tcl_DStringEndSublist(poutput);
  }
    
  if ( m && m->lpszSubject )
  {
    Tcl_DStringStartSublist(poutput);
    Tcl_DStringAppendElement(poutput, "Subject");
    if ( m->lpszSubject && m->lpszSubject[0] )
      Tcl_DStringAppendElement(poutput, m->lpszSubject);
    else
      Tcl_DStringAppendElement(poutput, "<No Subject>");
    Tcl_DStringEndSublist(poutput);
  }
  
  if ( m && m->lpszNoteText )
  {
    Tcl_DStringStartSublist(poutput);
    Tcl_DStringAppendElement(poutput, "Body");
    Tcl_DStringAppendElement(poutput, m->lpszNoteText);
    Tcl_DStringEndSublist(poutput);
  }
  if ( m && m->lpszDateReceived )
  {
    Tcl_DStringStartSublist(poutput);
    Tcl_DStringAppendElement(poutput, "Time");
    Tcl_DStringAppendElement(poutput, m->lpszDateReceived);
    Tcl_DStringEndSublist(poutput);
  }
    
  /* Originator */
  if ( m && m->lpOriginator )
  {
    Tcl_DStringStartSublist(poutput);
    Tcl_DStringAppendElement(poutput, "Originator");
    if ( m->lpOriginator )
    {
      if ( m->lpOriginator->lpszAddress && m->lpOriginator->lpszAddress[0] )
        Tcl_DStringAppendElement(poutput, m->lpOriginator->lpszAddress);
      else if (m->lpOriginator->lpszName && m->lpOriginator->lpszName )
        Tcl_DStringAppendElement(poutput, m->lpOriginator->lpszName);
      else
        Tcl_DStringAppendElement(poutput, "Unknown");
    }
    else
    {
      Tcl_DStringAppendElement(poutput, "Unknown");
    }
    Tcl_DStringEndSublist(poutput);
  }
    
  /* Recipients */
  if ( m && m->lpRecips )
  {
    Tcl_DStringStartSublist(poutput);
    Tcl_DStringAppendElement(poutput, "Recipients:");
    {
      unsigned j;
      for (j=0; j<m->nRecipCount; j++)
        Tcl_DStringAppendElement(poutput, m->lpRecips[j].lpszAddress);
    }
    Tcl_DStringEndSublist(poutput);
  }
    
  /* File names */
  if ( m && m->lpFiles )
  {
    /* To preserve all the information in the attachment, each attachment
    ** is represented as a 3-element list:
    ** First - The filename for the attachment (or copy)
    ** Second - The title of the attachment
    ** Third - a boolean representing whether the caller "owns" the attachment,
    **         and thus must delete it.
    */
    Tcl_DStringStartSublist(poutput);
    Tcl_DStringAppendElement(poutput, "Attachments:");
    {
      unsigned j;
      for (j=0; j<m->nFileCount; j++)
      {
        Tcl_DStringStartSublist(poutput);
        if ( m->lpFiles[j].lpszFileName )
          Tcl_DStringAppendElement(poutput, m->lpFiles[j].lpszFileName);
        else
          Tcl_DStringAppendElement(poutput, "");
        if ( m->lpFiles[j].lpszPathName )
          Tcl_DStringAppendElement(poutput, m->lpFiles[j].lpszPathName);
        else
          Tcl_DStringAppendElement(poutput, "");
        if ( responsibility != 0 )
          Tcl_DStringAppendElement(poutput, "0");
        else
          Tcl_DStringAppendElement(poutput, "1");
        Tcl_DStringEndSublist(poutput);
      }
    }
    Tcl_DStringEndSublist(poutput);
  }
    
}

static void FormatCmcDigestResult  ( Tcl_Interp *interp, CMC_message_summary FAR *pmsgref,
                                     Tcl_DString *poutput )
{
  char idstring[2*24+3];
  char *pidstring;
  int  size = sizeof(idstring);

  if ( (pidstring = CmcMsgRefToString(idstring, &size)) == 0 )
    pidstring = "Unknown";

#if DEBUG
  /* Debug */
  DumpMsgRef(&sessionstruct.cmc_msgref);
#endif
      
  /* OK. Now set up the output structure */
  /* For each element of the output, ensure it exists before copying it! */
  Tcl_DStringStartSublist(poutput);
  Tcl_DStringAppendElement(poutput, "id");
  Tcl_DStringAppendElement(poutput, pidstring);
  Tcl_DStringEndSublist(poutput);

  if ( pmsgref && pmsgref->subject )
  {
    Tcl_DStringStartSublist(poutput);
    Tcl_DStringAppendElement(poutput, "Subject");
    Tcl_DStringAppendElement(poutput, pmsgref->subject);
    Tcl_DStringEndSublist(poutput);
  }

  if ( pmsgref )
  {
    Tcl_DStringStartSublist(poutput);
    Tcl_DStringAppendElement(poutput, "Time");
    {
      char time_str[4+1+2+1+2+1+2+1+2+1];
      sprintf(time_str,"%4d/%02d/%02d %02d:%02d",
        1900+pmsgref->time_sent.year, pmsgref->time_sent.month+1,
        pmsgref->time_sent.day, pmsgref->time_sent.hour, pmsgref->time_sent.minute);
      Tcl_DStringAppendElement(poutput, time_str);
    }
    Tcl_DStringEndSublist(poutput);
  }
      
  /* Originator */
  if ( pmsgref && pmsgref->originator && pmsgref->originator->address )
  {
    Tcl_DStringStartSublist(poutput);
    Tcl_DStringAppendElement(poutput, "Originator");
    Tcl_DStringAppendElement(poutput, pmsgref->originator->address);
    Tcl_DStringEndSublist(poutput);
  }
      
  /* Size */
  Tcl_DStringStartSublist(poutput);
  Tcl_DStringAppendElement(poutput, "Size");
  {
    char size[14];
    if ( pmsgref && pmsgref->byte_length != CMC_LENGTH_UNKNOWN)
    {
      sprintf(size, "%ld", (long)pmsgref->byte_length);
      Tcl_DStringAppendElement(poutput, size);
    }
    else
      Tcl_DStringAppendElement(poutput, "Unknown");
  }
  Tcl_DStringEndSublist(poutput);
}                                     

static void FormatCmcCompleteResult ( Tcl_Interp *interp, CMC_message_summary FAR *pmsgref,
                                      CMC_message FAR *msg, Tcl_DString *poutput)
{
  if ( msg && msg->subject )
  {
    Tcl_DStringStartSublist(poutput );
    Tcl_DStringAppendElement(poutput, "Subject");
    Tcl_DStringAppendElement(poutput, msg->subject);
    Tcl_DStringEndSublist(poutput);
  }

  if ( msg && msg->text_note )
  {
    Tcl_DStringStartSublist(poutput);
    Tcl_DStringAppendElement(poutput, "Body");
    Tcl_DStringAppendElement(poutput, msg->text_note);
    Tcl_DStringEndSublist(poutput);
  }

  if (msg )
  {
    Tcl_DStringStartSublist(poutput);
    Tcl_DStringAppendElement(poutput, "Time");
    {
      char time_str[4+1+2+1+2+1+2+1+2+1];
      sprintf(time_str,"%4d/%02d/%02d %02d:%02d",
        1900+msg->time_sent.year, msg->time_sent.month+1,
        msg->time_sent.day, msg->time_sent.hour, msg->time_sent.minute);
      Tcl_DStringAppendElement(poutput, time_str);
    }
    Tcl_DStringEndSublist(poutput);
  }
      
  /* Originator */
  if ( msg )
  {
    Tcl_DStringStartSublist(poutput);
    Tcl_DStringAppendElement(poutput, "Originator");
    {
      /* Kind of brute force */
      int j;
      int k = -1;
      for (j=0; msg->recipients; j++ )
      {
        if ( msg->recipients[j].role == CMC_ROLE_ORIGINATOR )
        {
          k=j;
          break;
        }
        if ( msg->recipients[j].recip_flags & CMC_RECIP_LAST_ELEMENT )
          break;
      }
      if ( k == -1 )
        Tcl_DStringAppendElement(poutput, "Unknown");
      else
        Tcl_DStringAppendElement(poutput, msg->recipients[k].address);
    }
    Tcl_DStringEndSublist(poutput);
  }
      
  /* Recipients */
  if ( msg )
  {
    Tcl_DStringStartSublist(poutput);
    Tcl_DStringAppendElement(poutput, "Recipients");
    {
      int j;
      for (j=0; msg->recipients ; j++ )
      {
        if ( msg->recipients[j].role != CMC_ROLE_ORIGINATOR )
          Tcl_DStringAppendElement(poutput, msg->recipients[j].address);
        if ( msg->recipients[j].recip_flags & CMC_RECIP_LAST_ELEMENT )
          break;
      }
    }
    Tcl_DStringEndSublist(poutput);
  }

  /* Size: Only in the message reference, not in the message. */
  Tcl_DStringStartSublist(poutput);
  Tcl_DStringAppendElement(poutput, "Size");
  {
    char size[14];
    if ( pmsgref && pmsgref->byte_length != CMC_LENGTH_UNKNOWN)
    {
      sprintf(size, "%ld", (long)pmsgref->byte_length);
      Tcl_DStringAppendElement(poutput, size);
    }
    else
      Tcl_DStringAppendElement(poutput, "Unknown");
  }
  Tcl_DStringEndSublist(poutput);

  /* Attachments */
  Tcl_DStringStartSublist(poutput);
  Tcl_DStringAppendElement(poutput, "Attachments");
  {
    int j;
    /* To preserve all the information in the attachment, each attachment
    ** is represented as a 3-element list:
    ** First - The filename for the attachment (or copy)
    ** Second - The title of the attachment
    ** Third - a boolean representing whether the caller "owns" the attachment,
    **         and thus must delete it.
    */
    for (j=0; msg->attachments; j++)
    {
      Tcl_DStringStartSublist(poutput); /* Begin 3-element sublist */
      Tcl_DStringAppendElement(poutput, msg->attachments[j].attach_filename);
      Tcl_DStringAppendElement(poutput, msg->attachments[j].attach_title);
      if ( msg->attachments[j].attach_flags & CMC_ATT_APP_OWNS_FILE )
        Tcl_DStringAppendElement(poutput, "1");
      else
        Tcl_DStringAppendElement(poutput, "0");
      Tcl_DStringEndSublist(poutput);
      if (msg->attachments[j].attach_flags & CMC_ATT_LAST_ELEMENT)
        break;
    }
  }
  Tcl_DStringEndSublist(poutput);
}

/****************************************************************
** WinVersion returns an integer representing the current version
** of windows.
****************************************************************/
static long WinVersion(void)
{
    static OSVERSIONINFO osinfo;
    if ( osinfo.dwOSVersionInfoSize == 0 )
    {
      osinfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
      GetVersionEx(&osinfo); /* Should never fail--only failure is if size too small */
    }
    return osinfo.dwPlatformId;
}

/****************************************************************
** Initialization procedures
** These are the only public procedures in the file.
****************************************************************/
/* Initialization Procedures */
EXPORT(int,Mapimail_Init) (Tcl_Interp *interp)
{
  #if defined(USE_TCL_STUBS)
    Tcl_InitStubs(interp, TCL_VERSION, 0 );
  #endif

  Tcl_CreateCommand (interp, "mapimail", Mail, 0, 0);
  Tcl_PkgProvide (interp, "mapimail", version_string);
    
  return TCL_OK;
}

#if DEBUG

/****************************************************************/
/* Debug-only procedures */
/****************************************************************/

static void IAmHere(Tcl_Interp *interp, char *message)
{
  FILE *out;
  if ( (out = fopen("c:/tmp/out.log", "a")) != NULL )
  {
    fprintf(out, "Arrived at %s\n", message);
  }
  fflush(out);
  fclose (out);
}

static void DumpMsgRef(struct msgref *pmsgref)
{
  char *cp, *s;
  FILE *out;
  static char X[] = "0123456789ABCDEF";

  if ( (out = fopen("c:/tmp/out.log", "a")) != NULL )
  {
    if (pmsgref->length > 0 )
    {
      if ( (s = cp = calloc(2*pmsgref->length+3, sizeof(char)) ) != 0 )
      {
        int i;
        *cp++ = '0';
        *cp++ = 'x';
        for (i=0; i < pmsgref->length; i++)
        {
          *cp++ = X[(unsigned char)(pmsgref->string[i]) >> 4];
          *cp++ = X[(unsigned char)(pmsgref->string[i]) & 0xF];
        }
        fprintf(out, "reference dump is |%s| (length %ld)\n", s,
                (long)pmsgref->length);
        free(s);
      }
      else
      {
        fprintf(out, "Can't allocate memory for reference dump\n");
      }
    }
    else
    {
      fprintf(out, "Message reference length is 0\n");
    }
    fflush(out);
    fclose (out);
  }
}

#endif
